#! /usr/bin/env python
"""Utils for deduplication"""
import os
import time

import colrev.record

# pylint: disable=too-many-arguments


def console_duplicate_instance_label(
    record_pair: list,
    keys: list,
    manual: bool,
    curations_dupe_info: str,
    n_match: int,
    n_distinct: int,
    examples_buffer: list,
) -> str:
    """Convenience function for console duplicate labeling"""

    if manual:
        os.system("cls" if os.name == "nt" else "clear")
        colrev.record.Record.print_diff_pair(record_pair=record_pair, keys=keys)

    user_input = "unsure"
    if curations_dupe_info == "yes":
        user_input = "y"
        if manual:
            print(f"{n_match} positive, {n_distinct} negative")
            print("#")
            print("# curations_dupe_info: yes/duplicate")
            print("#")
            time.sleep(0.6)
    elif curations_dupe_info == "no":
        user_input = "n"
        if manual:
            print(f"{n_match} positive, {n_distinct} negative")
            print("#")
            print("# curations_dupe_info: no duplicate")
            print("#")
            time.sleep(0.6)
    else:
        if manual:
            print(f"{n_match} positive, {n_distinct} negative")

        if manual:
            valid_response = False
            user_input = ""

            while not valid_response:
                if examples_buffer:
                    prompt = (
                        "Duplicate? (y)es / (n)o / (u)nsure /"
                        + " (f)inished / (p)revious"
                    )
                    valid_responses = {"y", "n", "u", "f", "p"}
                else:
                    prompt = "Duplicate? (y)es / (n)o / (u)nsure / (f)inished"
                    valid_responses = {"y", "n", "u", "f"}

                print(prompt)
                user_input = input()
                if user_input in valid_responses:
                    valid_response = True
    return user_input



#! /usr/bin/env python
"""SearchSource: OpenCitations"""
from __future__ import annotations

import json
import typing
from dataclasses import dataclass
from pathlib import Path

import requests
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.crossref
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class OpenCitationsSearchSource(JsonSchemaMixin):
    """Performs a forward search based on OpenCitations
    Scope: all included papers with colrev_status in (rev_included, rev_synthesized)
    """

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "fwsearch_ref"
    search_type = colrev.settings.SearchType.FORWARD_SEARCH
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "OpenCitations forward search"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/open_citations_forward_search.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.review_manager = source_operation.review_manager
        self.crossref_connector = (
            colrev.ops.built_in.search_sources.crossref.CrossrefSearchSource(
                source_operation=source_operation
            )
        )
        self.__etiquette = self.crossref_connector.get_etiquette(
            review_manager=self.review_manager
        )

    @classmethod
    def get_default_source(cls) -> colrev.settings.SearchSource:
        """Get the default SearchSource settings"""
        return colrev.settings.SearchSource(
            endpoint="colrev.open_citations_forward_search",
            filename=Path("data/search/forward_search.bib"),
            search_type=colrev.settings.SearchType.FORWARD_SEARCH,
            search_parameters={
                "scope": {"colrev_status": "rev_included|rev_synthesized"}
            },
            comment="",
        )

    def __validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if "scope" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "Scope required in the search_parameters"
            )

        if (
            source.search_parameters["scope"]["colrev_status"]
            != "rev_included|rev_synthesized"
        ):
            raise colrev_exceptions.InvalidQueryException(
                "search_parameters/scope/colrev_status must be rev_included|rev_synthesized"
            )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def __fw_search_condition(self, *, record: dict) -> bool:
        if "doi" not in record:
            return False

        # rev_included/rev_synthesized required, but record not in rev_included/rev_synthesized
        if (
            "colrev_status" in self.search_source.search_parameters["scope"]
            and self.search_source.search_parameters["scope"]["colrev_status"]
            == "rev_included|rev_synthesized"
            and record["colrev_status"]
            not in [
                colrev.record.RecordState.rev_included,
                colrev.record.RecordState.rev_synthesized,
            ]
        ):
            return False

        return True

    def __get_forward_search_records(self, *, record_dict: dict) -> list:
        forward_citations = []

        url = f"https://opencitations.net/index/coci/api/v1/citations/{record_dict['doi']}"
        # headers = {"authorization": "YOUR-OPENCITATIONS-ACCESS-TOKEN"}
        headers: typing.Dict[str, str] = {}

        ret = requests.get(url, headers=headers, timeout=300)
        try:
            items = json.loads(ret.text)

            for doi in [x["citing"] for x in items]:
                retrieved_record = self.crossref_connector.query_doi(
                    doi=doi, etiquette=self.__etiquette
                )
                # if not crossref_query_return:
                #     raise colrev_exceptions.RecordNotFoundInPrepSourceException()
                retrieved_record.data["ID"] = retrieved_record.data["doi"]
                forward_citations.append(retrieved_record.data)
        except json.decoder.JSONDecodeError:
            self.review_manager.logger.info(
                f"Error retrieving citations from Opencitations for {record_dict['ID']}"
            )

        return forward_citations

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a forward search based on OpenCitations"""

        # pylint: disable=too-many-branches

        self.__validate_source()

        records = search_operation.review_manager.dataset.load_records_dict()

        if not records:
            print("No records imported. Cannot run forward search yet.")
            return

        forward_search_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        for record in records.values():
            if not self.__fw_search_condition(record=record):
                continue

            search_operation.review_manager.logger.info(
                f'Run forward search for {record["ID"]}'
            )

            new_records = self.__get_forward_search_records(record_dict=record)

            for new_record in new_records:
                new_record["fwsearch_ref"] = (
                    record["ID"] + "_forward_search_" + new_record["ID"]
                )
                new_record["cites_IDs"] = record["ID"]

                try:
                    forward_search_feed.set_id(record_dict=new_record)
                except colrev_exceptions.NotFeedIdentifiableException:
                    continue

                prev_record_dict_version = {}
                if new_record["ID"] in forward_search_feed.feed_records:
                    prev_record_dict_version = forward_search_feed.feed_records[
                        new_record["ID"]
                    ]

                added = forward_search_feed.add_record(
                    record=colrev.record.Record(data=new_record),
                )

                if added:
                    pass
                elif rerun:
                    # Note : only re-index/update
                    forward_search_feed.update_existing_record(
                        records=records,
                        record_dict=new_record,
                        prev_record_dict_version=prev_record_dict_version,
                        source=self.search_source,
                        update_time_variant_fields=rerun,
                    )

        forward_search_feed.save_feed_file()
        forward_search_feed.print_post_run_search_infos(records=records)

        if search_operation.review_manager.dataset.has_changes():
            search_operation.review_manager.create_commit(
                msg="Forward search", script_call="colrev search"
            )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for forward searches (OpenCitations)"""

        result = {"confidence": 0.0}

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint"""

        add_source = cls.get_default_source()
        operation.review_manager.settings.sources.append(add_source)

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for forward searches (OpenCitations)"""
        return record



#! /usr/bin/env python
"""Screening utilities"""
from __future__ import annotations

from typing import TYPE_CHECKING

import colrev.env.package_manager
import colrev.record
import colrev.settings

if TYPE_CHECKING:
    import colrev.ops.screen


__FULL_SCREEN_EXPLANATION = (
    "Explanation: Screening criteria can be used "
    + """to include or exclude records based on specific reasons.
Example:
    - short name        (short string): behavioral_treatment
    - criterion type    (select one)  : [i, e for inclusion_criterion OR exclusion_criterion]
    - explanation       (text)        : Include records reporting on behavioral treatment

Add a screening criterion [y,n]?"""
)

# Example 2:
#     - short name        : non_experimental_method
#     - criterion type    : exclusion_criterion
#     - explanation       : Exclude records reporting on non-experimental designs


def __get_add_screening_criterion_dialogue(*, screening_criteria: dict) -> str:
    if not screening_criteria:
        return __FULL_SCREEN_EXPLANATION
    return "Add another screening criterion [y,n]?"


def get_screening_criteria_from_user_input(
    *, screen_operation: colrev.ops.screen.Screen, records: dict
) -> dict:
    """Get the screening criteria from user input (initial setup)"""

    screening_criteria = screen_operation.review_manager.settings.screen.criteria
    if len(screening_criteria) == 0 and 0 == len(
        [
            r
            for r in records.values()
            if r["colrev_status"]
            in [
                colrev.record.RecordState.rev_included,
                colrev.record.RecordState.rev_excluded,
                colrev.record.RecordState.rev_synthesized,
            ]
        ]
    ):
        print()
        screening_criteria = {}
        while "y" == input(
            __get_add_screening_criterion_dialogue(
                screening_criteria=screening_criteria
            )
        ):
            short_name = input("Provide a short name: ")
            if input("Inclusion or exclusion criterion [i,e]?: ") == "i":
                criterion_type = colrev.settings.ScreenCriterionType.inclusion_criterion
            else:
                criterion_type = colrev.settings.ScreenCriterionType.exclusion_criterion
            explanation = input("Provide a short explanation: ")

            screening_criteria[short_name] = colrev.settings.ScreenCriterion(
                explanation=explanation, criterion_type=criterion_type, comment=""
            )
            print()

        screen_operation.set_screening_criteria(screening_criteria=screening_criteria)

    return screening_criteria



#!/usr/bin/env python3
"""Collection of utility functions"""
import operator
import pkgutil
import typing
import unicodedata
from enum import Enum
from functools import reduce
from pathlib import Path

from jinja2 import Environment
from jinja2 import FunctionLoader
from jinja2.environment import Template

import colrev.exceptions as colrev_exceptions


def retrieve_package_file(*, template_file: Path, target: Path) -> None:
    """Retrieve a file from the CoLRev package"""
    filedata = pkgutil.get_data("colrev", str(template_file))
    if not filedata:
        raise colrev_exceptions.RepoSetupError(f"{template_file} not available")

    target.parent.mkdir(exist_ok=True, parents=True)
    with open(target, "w", encoding="utf8") as file:
        file.write(filedata.decode("utf-8"))


def get_package_file_content(*, file_path: Path) -> typing.Union[bytes, None]:
    """Get the content of a file in the CoLRev package"""
    return pkgutil.get_data("colrev", str(file_path))


def inplace_change(*, filename: Path, old_string: str, new_string: str) -> None:
    """Replace a string in a file"""
    with open(filename, encoding="utf8") as file:
        content = file.read()
        if old_string not in content:
            return
    with open(filename, "w", encoding="utf8") as file:
        content = content.replace(old_string, new_string)
        file.write(content)


def get_template(*, template_path: str) -> Template:
    """Load a jinja template"""
    environment = Environment(
        loader=FunctionLoader(__load_jinja_template), autoescape=True
    )
    template = environment.get_template(template_path)
    return template


def __load_jinja_template(template_path: str) -> str:
    filedata_b = pkgutil.get_data("colrev", template_path)
    if filedata_b:
        filedata = filedata_b.decode("utf-8")
        filedata = filedata.replace("\n", "")
        filedata = filedata.replace("<br>", "\n")
        return filedata
    raise colrev_exceptions.RepoSetupError(f"{template_path} not available")


def remove_accents(*, input_str: str) -> str:
    """Replace the accents in a string"""

    def rmdiacritics(char: str) -> str:
        """
        Return the base character of char, by "removing" any
        diacritics like accents or curls and strokes and the like.
        """
        try:
            desc = unicodedata.name(char)
            cutoff = desc.find(" WITH ")
            if cutoff != -1:
                desc = desc[:cutoff]
                char = unicodedata.lookup(desc)
        except (KeyError, ValueError):
            pass  # removing "WITH ..." produced an invalid name
        return char

    try:
        nfkd_form = unicodedata.normalize("NFKD", input_str)
        wo_ac_list = [
            rmdiacritics(c) for c in nfkd_form if not unicodedata.combining(c)
        ]
        wo_ac = "".join(wo_ac_list)
    except ValueError:
        wo_ac = input_str
    return wo_ac


def percent_upper_chars(input_string: str) -> float:
    """Get the percentage of upper-case characters in a string"""
    return sum(map(str.isupper, input_string)) / len(input_string)


def custom_asdict_factory(data) -> dict:  # type: ignore
    """Custom asdict factory for (dataclass)object-to-dict conversion"""

    def convert_value(obj: object) -> object:
        if isinstance(obj, Enum):
            return obj.value
        if isinstance(obj, Path):
            return str(obj)
        if isinstance(obj, float):
            # Save 1.0 as 1 per default to avoid parsing issues
            # e.g., with the web ui
            if str(obj) == "1.0":
                return 1
        if isinstance(obj, list):
            if all(isinstance(el, Path) for el in obj):
                return [str(el) for el in obj]
        return obj

    return {k: convert_value(v) for k, v in data}


def load_complementary_material_keywords() -> list:
    """Load the list of keywords identifying complementary materials"""
    complementary_material_keywords = []
    filedata = get_package_file_content(
        file_path=Path("template/ops/complementary_material_keywords.txt")
    )
    if filedata:
        complementary_material_keywords = list(filedata.decode("utf-8").splitlines())

    return complementary_material_keywords


def load_complementary_material_strings() -> list:
    """Load the list of exact strings identifying complementary materials"""

    complementary_material_keywords = []
    filedata = get_package_file_content(
        file_path=Path("template/ops/complementary_material_strings.txt")
    )
    if filedata:
        complementary_material_keywords = list(filedata.decode("utf-8").splitlines())

    return complementary_material_keywords


def load_complementary_material_prefixes() -> list:
    """Load the list of prefixes identifying complementary materials"""

    complementary_material_keywords = []
    filedata = get_package_file_content(
        file_path=Path("template/ops/complementary_material_prefixes.txt")
    )
    if filedata:
        complementary_material_keywords = list(filedata.decode("utf-8").splitlines())

    return complementary_material_keywords


def dict_keys_exists(element: dict, *keys: str) -> bool:
    """Check if *keys (nested) exists in `element` (dict)."""
    if len(keys) == 0:
        raise AttributeError("keys_exists() expects at least two arguments, one given.")

    _element = element
    for key in keys:
        try:
            _element = _element[key]
        except KeyError:
            return False
    return True


def get_by_path(root: dict, items: typing.List[str]) -> typing.Any:
    """Access a nested object in root by item sequence."""

    return reduce(operator.getitem, items, root)


def set_by_path(root: dict, items: typing.List[str], value: typing.Any) -> None:
    """Set a value in a nested object in root by item sequence."""

    get_by_path(root, items[:-1])[items[-1]] = value


def dict_set_nested(root: dict, keys: typing.List[str], value: typing.Any) -> None:
    """Set dict value by nested key, this works on empty dict"""
    for key in keys[:-1]:
        root = root.setdefault(key, {})
    root[keys[-1]] = value



#! /usr/bin/env python
"""Prescreen based on CLI"""
from __future__ import annotations

import typing
from dataclasses import dataclass

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.record
import colrev.ui_cli.cli_colors as colors

if typing.TYPE_CHECKING:
    import colrev.ops.prescreen.Prescreen

# pylint: disable=too-few-public-methods


@zope.interface.implementer(
    colrev.env.package_manager.PrescreenPackageEndpointInterface
)
@dataclass
class CoLRevCLIPrescreen(JsonSchemaMixin):

    """CLI-based prescreen"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def __fun_cli_prescreen(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        prescreen_data: dict,
        split: list,
        stat_len: int,
        padding: int,
    ) -> bool:
        if prescreen_operation.review_manager.settings.prescreen.explanation == "":
            print(
                f"\n{colors.ORANGE}Provide a short explanation of the prescreen{colors.END} "
                "(why should particular papers be included?):"
            )
            print(
                'Example objective: "Include papers that focus on digital technology."'
            )
            prescreen_operation.review_manager.settings.prescreen.explanation = input(
                ""
            )
            prescreen_operation.review_manager.save_settings()
        else:
            print("\nIn the prescreen, the following process is followed:\n")
            print(
                "   "
                + prescreen_operation.review_manager.settings.prescreen.explanation
            )
            print()

        prescreen_operation.review_manager.logger.debug("Start prescreen")

        if 0 == stat_len:
            prescreen_operation.review_manager.logger.info("No records to prescreen")

        i, quit_pressed = 0, False
        for record_dict in prescreen_data["items"]:
            if len(split) > 0:
                if record_dict["ID"] not in split:
                    continue

            record = colrev.record.Record(data=record_dict)
            ret, inclusion_decision_str = "NA", "NA"
            i += 1

            print("\n\n")
            print(f"Record {i} (of {stat_len})\n")
            record.print_prescreen_record()

            while ret not in ["y", "n", "s", "q"]:
                ret = input(
                    "\nInclude this record "
                    "[enter y,n,s,q for yes, no, skip/decide later, quit-and-save]? "
                )
                if ret == "q":
                    quit_pressed = True
                elif ret == "s":
                    continue
                else:
                    inclusion_decision_str = ret.replace("y", "yes").replace("n", "no")

            if quit_pressed:
                prescreen_operation.review_manager.logger.info("Stop prescreen")
                break

            if ret == "s":
                continue

            inclusion_decision = "yes" == inclusion_decision_str
            prescreen_operation.prescreen(
                record=record,
                prescreen_inclusion=inclusion_decision,
                PAD=padding,
            )

        return i == stat_len

    def run_prescreen(
        self,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        records: dict,
        split: list,
    ) -> dict:
        """Prescreen records based on a cli"""

        if not split:
            split = []

        prescreen_data = prescreen_operation.get_data()
        stat_len = len(split) if len(split) > 0 else prescreen_data["nr_tasks"]
        padding = prescreen_data["PAD"]

        self.__fun_cli_prescreen(
            prescreen_operation=prescreen_operation,
            prescreen_data=prescreen_data,
            split=split,
            stat_len=stat_len,
            padding=padding,
        )

        # records = prescreen_operation.review_manager.dataset.load_records_dict()
        # prescreen_operation.review_manager.dataset.save_records_dict(records=records)
        prescreen_operation.review_manager.dataset.add_record_changes()

        # Note : currently, it is easier to create a commit in all cases.
        # Upon continuing the prescreen, the scope-based prescreen commits the changes,
        # which is misleading.
        # Users can still squash commits.
        # Note: originall: completed = self.__fun_cli_prescreen(...
        # if not completed:
        #     if input("Create commit (y/n)?") != "y":
        #         return records

        prescreen_operation.review_manager.create_commit(
            msg="Pre-screening (manual)", manual_author=True
        )
        return records



#! /usr/bin/env python
"""Checker for year-format."""
from __future__ import annotations

import re

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class YearFormatChecker:
    """The YearFormatChecker"""

    msg = "year-format"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the year-format checks"""

        if "year" not in record.data:
            return
        if record.data["year"] == "UNKNOWN":
            return

        if not re.match(r"^\d{4}$", record.data["year"]):
            record.add_masterdata_provenance_note(key="year", note=self.msg)
        else:
            record.remove_masterdata_provenance_note(key="year", note=self.msg)


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(YearFormatChecker(quality_model))



#! /usr/bin/env python
"""Colrev curated data as part of the data operations"""
from __future__ import annotations

import collections
import os
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.data


@zope.interface.implementer(colrev.env.package_manager.DataPackageEndpointInterface)
@dataclass
class ColrevCuration(JsonSchemaMixin):
    """CoLRev Curation"""

    settings: ColrevCurationSettings
    ci_supported: bool = True

    @dataclass
    class ColrevCurationSettings(
        colrev.env.package_manager.DefaultSettings, JsonSchemaMixin
    ):
        """Colrev Curation settings"""

        endpoint: str
        version: str
        curation_url: str
        curated_masterdata: bool
        masterdata_restrictions: dict
        curated_fields: list

    settings_class = ColrevCurationSettings

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
        # Set default values (if necessary)
        if "version" not in settings:
            settings["version"] = "0.1"

        self.settings = self.settings_class.load_settings(data=settings)

        self.data_operation = data_operation

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_source = {
            "endpoint": "colrev.colrev_curation",
            "version": "0.1",
            "curation_url": "TODO",
            "curated_masterdata": True,
            "masterdata_restrictions": {
                1900: {
                    "ENTRYTYPE": "article",
                    "journal": "TODO",
                    "volume": True,
                    "number": True,
                }
            },
            "curated_fields": ["doi", "url"],
        }

        operation.review_manager.settings.data.data_package_endpoints.append(add_source)

    def __get_stats(
        self,
        *,
        records: dict,
        sources: list,
    ) -> dict:
        # pylint: disable=too-many-branches

        stats: dict = {}
        for record_dict in records.values():
            r_status = str(record_dict["colrev_status"])
            if r_status == "rev_prescreen_excluded":
                continue
            if record_dict[
                "colrev_status"
            ] in colrev.record.RecordState.get_post_x_states(
                state=colrev.record.RecordState.md_processed
            ):
                r_status = str(colrev.record.RecordState.md_processed)
            elif record_dict[
                "colrev_status"
            ] in colrev.record.RecordState.get_post_x_states(
                state=colrev.record.RecordState.pdf_prepared
            ):
                r_status = str(colrev.record.RecordState.pdf_prepared)
            else:
                r_status = str(colrev.record.RecordState.md_imported)

            if "journal" in record_dict:
                key = (
                    f"{record_dict.get('year','-')}-"
                    f"{record_dict.get('volume','-')}-"
                    f"{record_dict.get('number','-')}"
                )
            elif "booktitle" in record_dict:
                key = record_dict.get("year", "-")
            else:
                self.data_operation.review_manager.logger.error(
                    f"TOC not supported: {record_dict}"
                )
                continue
            if not all(
                source in [o.split("/")[0] for o in record_dict["colrev_origin"]]
                for source in sources
            ):
                if key in stats:
                    if "all_merged" in stats[key]:
                        stats[key]["all_merged"] = "NO"
                    else:
                        stats[key]["all_merged"] = "NO"
                else:
                    stats[key] = {"all_merged": "NO"}

            for origin in record_dict["colrev_origin"]:
                source = origin.split("/")[0]
                if key in stats:
                    if source in stats[key]:
                        if r_status in stats[key][source]:
                            stats[key][source][r_status] += 1
                        else:
                            stats[key][source][r_status] = 1
                    else:
                        stats[key][source] = {r_status: 1}
                else:
                    stats[key] = {source: {r_status: 1}}
        return stats

    def __get_stats_markdown_table(self, *, stats: dict, sources: list) -> str:
        cell_width = 16
        sources += ["all_merged"]
        output = "|TOC".ljust(cell_width - 1, " ") + "|"
        sub_header_lines = "|".ljust(cell_width - 1, "-") + "|"
        for source in sources:
            output += f"{source}".ljust(cell_width, " ") + "|"
            sub_header_lines += "".ljust(cell_width, "-") + "|"

        output += "\n" + sub_header_lines
        ordered_stats = collections.OrderedDict(sorted(stats.items(), reverse=True))

        output += self.__get_stats_markdown_table_cell(
            ordered_stats=ordered_stats, sources=sources, cell_width=cell_width
        )

        output += "\n\nLegend: *md_imported*, md_processed, **pdf_prepared**"
        return output

    def __get_stats_markdown_table_cell(
        self, *, ordered_stats: collections.OrderedDict, sources: list, cell_width: int
    ) -> str:
        output = ""
        for key, row in ordered_stats.items():
            output += f"\n|{key}".ljust(cell_width, " ") + "|"
            for source in sources:
                if source != "all_merged":
                    if source in row:
                        cell_text = ""
                        if "md_imported" in row[source]:
                            cell_text += f"*{row[source]['md_imported']}*,"
                        else:
                            cell_text += "-,"
                        if "md_processed" in row[source]:
                            cell_text += f"{row[source]['md_processed']}"
                        else:
                            cell_text += "-"
                        if "pdf_prepared" in row[source]:
                            cell_text += f",**{row[source]['pdf_prepared']}**"
                        else:
                            cell_text += ",-"

                        output += cell_text.rjust(cell_width, " ") + "|"
                    else:
                        output += "-".rjust(cell_width, " ") + "|"
                else:
                    output += row.get("all_merged", "").rjust(cell_width, " ") + "|"
        return output

    def __update_table_in_readme(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        markdown_output: str,
    ) -> None:
        table_summary_tag = "<!-- TABLE_SUMMARY -->"
        readme_path = review_manager.readme
        with open(readme_path, "r+b") as file:
            appended = False
            seekpos = file.tell()
            line = file.readline()
            while line:
                if table_summary_tag.encode("utf-8") in line:
                    line = file.readline()
                    while (
                        b"Legend: *md_imported*, md_processed" not in line and line
                    ):  # replace: drop the current record
                        line = file.readline()

                    remaining = file.read()
                    file.seek(seekpos)
                    file.write(table_summary_tag.encode("utf-8"))
                    file.write(b"\n\n")
                    file.write(markdown_output.encode("utf-8"))
                    file.write(b"\n")
                    seekpos = file.tell()
                    file.flush()
                    os.fsync(file)
                    file.write(remaining)
                    file.truncate()  # if the replacement is shorter...
                    file.seek(seekpos)
                    appended = True

                seekpos = file.tell()
                line = file.readline()
            if not appended:
                file.write(b"\n")
                file.write(table_summary_tag.encode("utf-8"))
                file.write(b"\n\n")
                file.write(markdown_output.encode("utf-8"))
                file.write(b"\n")

    def __update_stats_in_readme(
        self,
        *,
        records: dict,
        review_manager: colrev.review_manager.ReviewManager,
        silent_mode: bool,
    ) -> None:
        if not silent_mode:
            review_manager.logger.info("Calculate statistics for readme")

        # alternatively: get sources from search_sources.filename (name/stem?)
        sources = []
        for record_dict in records.values():
            for origin in record_dict["colrev_origin"]:
                source = origin.split("/")[0]
                if source not in sources:
                    sources.append(source)

        stats = self.__get_stats(records=records, sources=sources)
        markdown_output = self.__get_stats_markdown_table(stats=stats, sources=sources)
        self.__update_table_in_readme(
            review_manager=review_manager, markdown_output=markdown_output
        )
        review_manager.dataset.add_changes(path=review_manager.README_RELATIVE)

    def __source_comparison(self, *, silent_mode: bool) -> None:
        """Exports a table to support analyses of records that are not
        in all sources (for curated repositories)"""

        self.data_operation.review_manager.dedupe_dir.mkdir(exist_ok=True, parents=True)
        source_comparison_xlsx = self.data_operation.review_manager.dedupe_dir / Path(
            "source_comparison.xlsx"
        )

        source_filenames = [
            str(x.filename) for x in self.data_operation.review_manager.settings.sources
        ]
        if not silent_mode:
            print("sources: " + ",".join([str(x) for x in source_filenames]))

        records = self.data_operation.review_manager.dataset.load_records_dict()
        records = {
            k: v
            for k, v in records.items()
            if not all(x in ";".join(v["colrev_origin"]) for x in str(source_filenames))
        }
        if len(records) == 0:
            if not silent_mode:
                print("No records unmatched")
            return

        for record in records.values():
            origins = record["colrev_origin"]
            for source_filename in source_filenames:
                if not any(source_filename in origin for origin in origins):
                    record[source_filename] = ""
                else:
                    record[source_filename] = [
                        origin for origin in origins if source_filename in origin
                    ][0]
            record["merge_with"] = ""

        records_df = pd.DataFrame.from_records(list(records.values()))
        records_df.to_excel(source_comparison_xlsx, index=False)
        if not silent_mode:
            print(f"Exported {source_comparison_xlsx}")

    def update_data(
        self,
        data_operation: colrev.ops.data.Data,
        records: dict,
        synthesized_record_status_matrix: dict,  # pylint: disable=unused-argument
        silent_mode: bool,
    ) -> None:
        """Update the CoLRev curation"""

        if self.settings.curated_masterdata:
            self.__update_stats_in_readme(
                records=records,
                review_manager=data_operation.review_manager,
                silent_mode=silent_mode,
            )
            self.__source_comparison(silent_mode=silent_mode)

    def update_record_status_matrix(
        self,
        data_operation: colrev.ops.data.Data,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        for item in synthesized_record_status_matrix:
            synthesized_record_status_matrix[item][endpoint_identifier] = True

    def get_advice(
        self,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        records = review_manager.dataset.load_records_dict()
        advice = {
            "msg": "TODO (add curation-specific advice...)",
            "detailed_msg": "TODO",
        }

        records_missing_languages = [
            r["ID"] for r in records.values() if "language" not in r
        ]
        if records_missing_languages:
            advice = {
                "msg": "Curation: Add language field to all records",
                "detailed_msg": "records missing language field: "
                + f"({','.join(records_missing_languages)})",
            }

        identical_colrev_ids: typing.Dict[str, list] = {}
        non_identifiable_records = []
        for record_dict in records.values():
            try:
                if record_dict[
                    "colrev_status"
                ] not in colrev.record.RecordState.get_post_x_states(
                    state=colrev.record.RecordState.md_prepared
                ):
                    continue
                if (
                    record_dict["colrev_status"]
                    == colrev.record.RecordState.rev_prescreen_excluded
                ):
                    continue
                cid = colrev.record.Record(data=record_dict).create_colrev_id(
                    assume_complete=True
                )
                if cid in identical_colrev_ids:
                    identical_colrev_ids[cid] = identical_colrev_ids[cid] + [
                        record_dict["ID"]
                    ]
                else:
                    identical_colrev_ids[cid] = [record_dict["ID"]]
            except colrev_exceptions.NotEnoughDataToIdentifyException:
                non_identifiable_records.append(record_dict["ID"])

        identical_colrev_ids = {
            k: v for k, v in identical_colrev_ids.items() if len(v) > 1
        }
        if identical_colrev_ids:
            advice = {
                "msg": "Curation: resolve records with identical colrev_ids:\n      - "
                + "\n      - ".join(",".join(x) for x in identical_colrev_ids.values()),
                "detailed_msg": "records missing language field: ",
            }

        if not self.settings.masterdata_restrictions:
            advice = {
                "msg": "Curation: masterdata_restrictions not set. See "
                + "https://github.com/CoLRev-Environment/colrev/blob/main/"
                + "colrev/ops/built_in/data/colrev_curation.md for details.",
                "detailed_msg": "records missing language field: ",
            }

        return advice



#!/usr/bin/env python3
"""Hook for reporting in CoLRev projects"""
import sys
from pathlib import Path

import colrev.review_manager


def main() -> int:
    """Main entrypoint for the reporting"""

    print(sys.argv)
    msgfile = Path(sys.argv[1])

    review_manager = colrev.review_manager.ReviewManager()
    ret = review_manager.report(msg_file=msgfile)

    print(ret["msg"])

    return ret["status"]


if __name__ == "__main__":
    raise SystemExit(main())



#! /usr/bin/env python
"""Checker for missing fields."""
from __future__ import annotations

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class MissingFieldChecker:
    """The MissingFieldChecker"""

    # Based on https://en.wikipedia.org/wiki/BibTeX
    record_field_requirements = {
        "article": ["author", "title", "journal", "year", "volume", "number"],
        "inproceedings": ["author", "title", "booktitle", "year"],
        "incollection": ["author", "title", "booktitle", "publisher", "year"],
        "inbook": ["author", "title", "chapter", "publisher", "year"],
        "proceedings": ["booktitle", "editor", "year"],
        "conference": ["booktitle", "editor", "year"],
        "book": ["author", "title", "publisher", "year"],
        "phdthesis": ["author", "title", "school", "year"],
        "bachelorthesis": ["author", "title", "school", "year"],
        "thesis": ["author", "title", "school", "year"],
        "masterthesis": ["author", "title", "school", "year"],
        "techreport": ["author", "title", "institution", "year"],
        "unpublished": ["title", "author", "year"],
        "misc": ["author", "title", "year"],
        "software": ["author", "title", "url"],
        "online": ["author", "title", "url"],
        "other": ["author", "title", "year"],
    }
    """Fields requirements for respective ENTRYTYPE"""

    # book, inbook: author <- editor

    msg = "missing"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model
        self.curation_restrictions = self.__load_curation_restrictions()
        self.__complementary_material_strings = (
            colrev.env.utils.load_complementary_material_strings()
        )
        self.__complementary_material_keywords = (
            colrev.env.utils.load_complementary_material_keywords()
        )

    def __load_curation_restrictions(self) -> dict:
        curation_restrictions = {}
        curated_endpoints = [
            x
            for x in self.quality_model.review_manager.settings.data.data_package_endpoints
            if x["endpoint"] == "colrev.colrev_curation"
        ]
        if curated_endpoints:
            curated_endpoint = curated_endpoints[0]
            curation_restrictions = curated_endpoint.get("masterdata_restrictions", {})
        return curation_restrictions

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the missing-field checks"""

        if self.curation_restrictions:
            self.__check_completeness_curated_masterdata(record=record)
        else:
            self.__check_completeness(record=record)

        self.__ignore_exceptions(record=record)
        self.__check_forthcoming(record=record)

        if not self.__has_missing_fields(record=record):
            record.set_masterdata_complete(
                source="update_masterdata_provenance",
                masterdata_repository=False,
                replace_source=False,
            )

    def __ignore_exceptions(self, *, record: colrev.record.Record) -> None:
        # No authors for complementary materials or errata required
        if (
            any(
                x == record.data.get("title", "UNKNOWN").lower()
                for x in self.__complementary_material_strings
            )
            or any(
                x in record.data.get("title", "UNKNOWN").lower()
                for x in self.__complementary_material_keywords
            )
            or any(
                x in record.data.get("title", "UNKNOWN").lower()
                for x in ["errata", "erratum", "corrigendum"]
            )
        ):
            if "author" in record.data["colrev_masterdata_provenance"] and any(
                x == "missing"
                for x in record.data["colrev_masterdata_provenance"]["author"][
                    "note"
                ].split(",")
            ):
                record.remove_field(
                    key="author", not_missing_note=True, source="ignored_exception"
                )

    def __has_missing_fields(self, *, record: colrev.record.Record) -> bool:
        if any(
            "missing" in x["note"].split(",")
            for x in record.data["colrev_masterdata_provenance"].values()
        ):
            return True
        return False

    def __check_forthcoming(self, *, record: colrev.record.Record) -> None:
        if record.data.get("year", "") != "forthcoming":
            record.remove_masterdata_provenance_note(key="volume", note="forthcoming")
            record.remove_masterdata_provenance_note(key="number", note="forthcoming")
            return
        source = "NA"
        if "year" in record.data["colrev_masterdata_provenance"]:
            source = record.data["colrev_masterdata_provenance"]["year"]["source"]
        if record.data.get("volume", "") in ["", "UNKNOWN"]:
            record.remove_masterdata_provenance_note(key="volume", note="missing")
            record.remove_masterdata_provenance_note(key="volume", note="not-missing")
            record.add_masterdata_provenance(
                key="volume", source=source, note="forthcoming"
            )
        if record.data.get("number", "") in ["", "UNKNOWN"]:
            record.remove_masterdata_provenance_note(key="number", note="missing")
            record.remove_masterdata_provenance_note(key="number", note="not-missing")
            record.add_masterdata_provenance(
                key="number", source=source, note="forthcoming"
            )

    def __check_completeness(
        self, *, record: colrev.record.Record, curation_restrictions: bool = False
    ) -> None:
        required_fields_keys = self.record_field_requirements["other"]
        if record.data["ENTRYTYPE"] in self.record_field_requirements:
            required_fields_keys = self.record_field_requirements[
                record.data["ENTRYTYPE"]
            ]
        else:
            raise colrev_exceptions.MissingRecordQualityRuleSpecification(
                msg=f"Missing record_field_requirements for {record.data['ENTRYTYPE']}"
            )

        if curation_restrictions:
            source = "colrev_curation.curation_restrictions"
        else:
            source = "generic_field_requirements"

        for required_fields_key in required_fields_keys:
            not_missing_note = False
            if required_fields_key in record.data["colrev_masterdata_provenance"]:
                if (
                    "not-missing"
                    in record.data["colrev_masterdata_provenance"][required_fields_key][
                        "note"
                    ]
                ):
                    not_missing_note = True

            if (
                record.data.get(required_fields_key, "UNKNOWN") == "UNKNOWN"
                and not not_missing_note
            ):
                record.update_field(
                    key=required_fields_key,
                    value="UNKNOWN",
                    source=source,
                    note=self.msg,
                    append_edit=False,
                )

        for required_field in ["author", "title", "year"]:
            if required_field in record.data:
                continue
            colrev.record.Record(data=record.data).add_masterdata_provenance(
                key=required_field,
                source=source,
                note=self.msg,
            )

    def __check_completeness_curated_masterdata(
        self, *, record: colrev.record.Record
    ) -> None:
        applicable_curation_restrictions = self.__get_applicable_curation_restrictions(
            record=record
        )

        self.__check_completeness(record=record, curation_restrictions=True)

        if "volume" in applicable_curation_restrictions:
            if (
                applicable_curation_restrictions["volume"]
                and "volume" not in record.data
            ):
                colrev.record.Record(data=record.data).add_masterdata_provenance(
                    key="volume",
                    source="colrev_curation.curation_restrictions",
                    note=self.msg,
                )
            else:
                colrev.record.Record(
                    data=record.data
                ).remove_masterdata_provenance_note(
                    key="volume",
                    note=self.msg,
                )

        if "number" in applicable_curation_restrictions:
            if (
                applicable_curation_restrictions["number"]
                and "number" not in record.data
            ):
                colrev.record.Record(data=record.data).add_masterdata_provenance(
                    key="number",
                    source="colrev_curation.curation_restrictions",
                    note=self.msg,
                )
            else:
                colrev.record.Record(
                    data=record.data
                ).remove_masterdata_provenance_note(
                    key="volume",
                    note=self.msg,
                )

        self.apply_curation_restrictions(record=record)

    def __get_applicable_curation_restrictions(
        self, *, record: colrev.record.Record
    ) -> dict:
        """Get the applicable curation restrictions"""

        if not str(record.data.get("year", "NA")).isdigit():
            return {}

        start_year_values = list(self.curation_restrictions.keys())
        year_index_diffs = [
            int(record.data["year"]) - int(x) for x in start_year_values
        ]
        year_index_diffs = [x if x >= 0 else 2000 for x in year_index_diffs]

        if not year_index_diffs:
            return {}

        index_min = min(range(len(year_index_diffs)), key=year_index_diffs.__getitem__)
        applicable_curation_restrictions = self.curation_restrictions[
            start_year_values[index_min]
        ]
        return applicable_curation_restrictions

    # Extract the restrictions to a qm-utility function?
    def apply_curation_restrictions(self, *, record: colrev.record.Record) -> None:
        """Apply the curation restrictions"""
        applicable_curation_restrictions = self.__get_applicable_curation_restrictions(
            record=record
        )
        if "ENTRYTYPE" in applicable_curation_restrictions:
            if applicable_curation_restrictions["ENTRYTYPE"] != record.data.get(
                "ENTRYTYPE", ""
            ):
                try:
                    record.change_entrytype(
                        new_entrytype=applicable_curation_restrictions["ENTRYTYPE"],
                        qm=self.quality_model,
                    )
                except colrev_exceptions.MissingRecordQualityRuleSpecification as exc:
                    print(exc)

        for field in ["journal", "booktitle"]:
            if field not in applicable_curation_restrictions:
                continue
            if applicable_curation_restrictions[field] == record.data.get(field, ""):
                continue
            record.update_field(
                key=field,
                value=applicable_curation_restrictions[field],
                source="colrev_curation.curation_restrictions",
            )
        if (
            "number" in applicable_curation_restrictions
            and not applicable_curation_restrictions["number"]
            and "number" in record.data
        ):
            record.remove_field(
                key="number",
                not_missing_note=True,
                source="colrev_curation.curation_restrictions",
            )


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(MissingFieldChecker(quality_model))



#!/usr/bin/env python3
"""Template for a custom Prep PackageEndpoint"""
from __future__ import annotations

from typing import TYPE_CHECKING

import zope.interface
from dacite import from_dict

import colrev.operation

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
class CustomPrep:
    """Class for custom prep scripts"""

    source_correction_hint = "check with the developer"
    always_apply_changes = True
    settings_class = colrev.env.package_manager.DefaultSettings

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = from_dict(data_class=self.settings_class, data=settings)

    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        record: colrev.record.Record,
    ) -> colrev.record.Record:
        """Update record (metadata)"""

        if "journal" in record.data:
            if record.data["journal"] == "MISQ":
                record.update_field(
                    key="journal", value="MIS Quarterly", source="custom_prep"
                )

        return record



#! /usr/bin/env python
"""Service to detect languages and handle language codes"""
from __future__ import annotations

import pycountry
from lingua.builder import LanguageDetectorBuilder

import colrev.exceptions as colrev_exceptions
import colrev.record


class LanguageService:
    """Service to detect languages and handle language codes"""

    __eng_false_negatives = ["editorial", "introduction"]

    def __init__(self) -> None:
        # Note : Lingua is tested/evaluated relative to other libraries:
        # https://github.com/pemistahl/lingua-py
        # It performs particularly well for short strings (single words/word pairs)
        # The langdetect library is non-deterministic, especially for short strings
        # https://pypi.org/project/langdetect/

        self.__lingua_language_detector = (
            LanguageDetectorBuilder.from_all_languages_with_latin_script().build()
        )

        # Language formats: ISO 639-1 standard language codes
        # https://pypi.org/project/langcodes/
        # https://github.com/flyingcircusio/pycountry

        self.__lang_code_mapping = {}
        for country in pycountry.languages:
            self.__lang_code_mapping[country.name.lower()] = country.alpha_3

    def compute_language(self, *, text: str) -> str:
        """Compute the most likely language code"""

        if text.lower() in self.__eng_false_negatives:
            return "eng"

        language = self.__lingua_language_detector.detect_language_of(text)
        if language:
            return language.iso_code_639_3.name.lower()
        return ""

    def compute_language_confidence_values(self, *, text: str) -> list:
        """Computes the most likely languages of a string and their language codes"""

        if text.lower() in self.__eng_false_negatives:
            return [("eng", 1.0)]

        predictions = (
            self.__lingua_language_detector.compute_language_confidence_values(
                text=text
            )
        )
        predictions_unified = []
        for lang, conf in predictions:
            predictions_unified.append((lang.iso_code_639_3.name.lower(), conf))

        return predictions_unified

    def validate_iso_639_3_language_codes(self, *, lang_code_list: list) -> None:
        """Validates whether a list of language codes complies with the ISO 639-3 standard"""

        assert isinstance(lang_code_list, list)

        invalid_language_codes = [x for x in lang_code_list if 3 != len(x)]
        if invalid_language_codes:
            raise colrev_exceptions.InvalidLanguageCodeException(
                invalid_language_codes=invalid_language_codes
            )

    def unify_to_iso_639_3_language_codes(
        self, *, record: colrev.record.Record
    ) -> None:
        """Unifies a language_code string to the ISO 639-3 standard"""

        if "language" not in record.data:
            return

        if record.data["language"].lower() in ["en"]:
            record.data["language"] = "eng"

        elif record.data["language"].lower() in ["fr"]:
            record.data["language"] = "fra"

        elif record.data["language"].lower() in ["ar"]:
            record.data["language"] = "ara"

        elif record.data["language"].lower() in ["de"]:
            record.data["language"] = "deu"

        if len(record.data["language"]) != 3:
            if record.data["language"].lower() in self.__lang_code_mapping:
                record.data["language"] = self.__lang_code_mapping[
                    record.data["language"].lower()
                ]

        self.validate_iso_639_3_language_codes(lang_code_list=[record.data["language"]])



#! /usr/bin/env python
"""Checker for pubmedid-not-matching-pattern."""
from __future__ import annotations

import re

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class PubmedIDPatternChecker:
    """The PubmedIDPatternChecker"""

    msg = "pubmedid-not-matching-pattern"

    __PMID_REGEX = r"^\d{1,8}(\.\d)?$"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the pubmedid-not-matching-pattern checks"""

        if "colrev.pubmed.pubmedid" not in record.data:
            return

        if not re.match(self.__PMID_REGEX, record.data["colrev.pubmed.pubmedid"]):
            record.add_masterdata_provenance_note(
                key="colrev.pubmed.pubmedid", note=self.msg
            )
        else:
            record.remove_masterdata_provenance_note(
                key="colrev.pubmed.pubmedid", note=self.msg
            )


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(PubmedIDPatternChecker(quality_model))



#! /usr/bin/env python
"""SearchSource: PsycINFO"""
from __future__ import annotations

import re
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin
from rispy import BaseParser
from rispy.config import LIST_TYPE_TAGS
from rispy.config import TAG_KEY_MAPPING

import colrev.env.package_manager
import colrev.ops.load_utils_ris
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


class PsycInfoRISParser(BaseParser):
    """Parser for Psycinfo RIS files."""

    START_TAG = "TY"
    IGNORE = ["FN", "VR", "EF"]
    PATTERN = r"^[A-Z][A-Z0-9]+ |^ER\s?|^EF\s?"
    mapping = deepcopy(TAG_KEY_MAPPING)
    # mapping["A1"] = "authors"
    mapping["PM"] = "pubmedid"
    # mapping["T1"] = "primary_title"
    # mapping["JF"] = "secondary_title"
    DEFAULT_MAPPING = mapping
    DEFAULT_LIST_TAGS = LIST_TYPE_TAGS

    def get_content(self, line: str) -> str:
        "Get the content from a line."
        return line[line.find(" - ") + 2 :].strip()

    def is_header(self, line: str) -> bool:
        "Check whether the line is a header element"
        return not re.match("[A-Z0-9]+  - ", line)


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class PsycINFOSearchSource(JsonSchemaMixin):
    """SearchSource for PsycINFO"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "url"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.oni
    short_name = "PsycInfo (APA)"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/psycinfo.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for PsycINFO"""

        result = {"confidence": 0.1}

        # Note : no features in bib file for identification

        if data.startswith(
            "Provider: American Psychological Association\nDatabase: PsycINFO"
        ):
            result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of Psycinfo"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def __ris_fixes(self, *, entries: dict) -> None:
        for entry in entries:
            if "alternate_title3" in entry and entry["type_of_reference"] in ["JOUR"]:
                entry["secondary_title"] = entry.pop("alternate_title3")
            if "publication_year" in entry:
                entry["year"] = entry.pop("publication_year")
            if "first_authors" in entry and "authors" not in entry:
                entry["authors"] = entry.pop("first_authors")

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".ris":
            ris_loader = colrev.ops.load_utils_ris.RISLoader(
                load_operation=load_operation, source=self.search_source
            )
            ris_entries = ris_loader.load_ris_entries(
                filename=self.search_source.filename, ris_parser=PsycInfoRISParser
            )
            self.__ris_fixes(entries=ris_entries)
            records = ris_loader.convert_to_records(entries=ris_entries)
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for PsycINFO"""

        record.rename_field(
            key="colrev.psycinfo.pubmedid", new_key="colrev.pubmed.pubmedid"
        )

        return record



#! /usr/bin/env python
"""CoLRev pdf_get operation: Get PDF documents."""
from __future__ import annotations

import os
import shutil
import typing
from glob import glob
from multiprocessing.pool import ThreadPool as Pool
from pathlib import Path

import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record
import colrev.ui_cli.cli_colors as colors


class PDFGet(colrev.operation.Operation):
    """Get the PDFs"""

    to_retrieve: int
    retrieved: int
    not_retrieved: int

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.pdf_get,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.package_manager = self.review_manager.get_package_manager()

        self.review_manager.pdf_dir.mkdir(exist_ok=True, parents=True)

        self.filepath_directory_pattern = ""
        pdf_endpoints = [
            s
            for s in self.review_manager.settings.sources
            if s.endpoint == "colrev.pdfs_dir"
        ]
        if pdf_endpoints:
            self.filepath_directory_pattern = (
                pdf_endpoints[0].search_parameters["scope"].get("subdir_pattern", {})
            )

    def copy_pdfs_to_repo(self) -> None:
        """Copy the PDFs to the repository"""
        self.review_manager.logger.info("Copy PDFs to dir")
        records = self.review_manager.dataset.load_records_dict()

        for record_dict in records.values():
            if "file" not in record_dict:
                continue
            fpath = Path(record_dict["file"])
            new_fpath = fpath.absolute()
            if fpath.is_symlink():
                linked_file = fpath.resolve()
                if linked_file.is_file():
                    fpath.unlink()
                    shutil.copyfile(linked_file, new_fpath)
                    self.review_manager.logger.info(
                        f' {colors.GREEN}copied PDF for {record_dict["ID"]} {colors.END}'
                    )

            elif new_fpath.is_file() and self.review_manager.verbose_mode:
                self.review_manager.logger.info(
                    f'No need to copy PDF - already exits ({record_dict["ID"]})'
                )

    def link_pdf(self, *, record: colrev.record.Record) -> colrev.record.Record:
        """Link the PDF in its record (should be {ID}.pdf)"""

        pdf_filepath = self.review_manager.PDF_DIR_RELATIVE / Path(
            f"{record.data['ID']}.pdf"
        )
        if pdf_filepath.is_file() and str(pdf_filepath) != record.data.get(
            "file", "NA"
        ):
            record.update_field(key="file", value=str(pdf_filepath), source="link_pdf")
            self.import_pdf(record=record)
            if (
                colrev.record.RecordState.rev_prescreen_included
                == record.data["colrev_status"]
            ):
                record.set_status(target_state=colrev.record.RecordState.pdf_imported)

        return record

    def get_target_filepath(self, *, record: colrev.record.Record) -> Path:
        """Get the target filepath for a PDF"""

        if self.filepath_directory_pattern == "year":
            target_filepath = self.review_manager.PDF_DIR_RELATIVE / Path(
                f"{record.data.get('year', 'no_year')}/{record.data['ID']}.pdf"
            )

        elif self.filepath_directory_pattern == "volume_number":
            if "volume" in record.data and "number" in record.data:
                target_filepath = self.review_manager.PDF_DIR_RELATIVE / Path(
                    f"{record.data['volume']}/{record.data['number']}/{record.data['ID']}.pdf"
                )

            if "volume" in record.data and "number" not in record.data:
                target_filepath = self.review_manager.PDF_DIR_RELATIVE / Path(
                    f"{record.data['volume']}/{record.data['ID']}.pdf"
                )
        else:
            target_filepath = self.review_manager.PDF_DIR_RELATIVE / Path(
                f"{record.data['ID']}.pdf"
            )

        return target_filepath

    def import_pdf(self, *, record: colrev.record.Record) -> None:
        """Import a file (PDF) and copy/symlink it"""
        # self.review_manager.pdf_dir.mkdir(exist_ok=True)
        # new_fp = self.review_manager.PDF_DIR_RELATIVE / Path(record.data["ID"] + ".pdf").name
        new_fp = self.get_target_filepath(record=record)
        original_fp = Path(record.data["file"])

        if new_fp != original_fp and not new_fp.is_file():
            new_fp.parents[0].mkdir(exist_ok=True, parents=True)
            if (
                colrev.settings.PDFPathType.symlink
                == self.review_manager.settings.pdf_get.pdf_path_type
            ):
                new_fp.symlink_to(original_fp)
            elif (
                colrev.settings.PDFPathType.copy
                == self.review_manager.settings.pdf_get.pdf_path_type
            ):
                shutil.copyfile(original_fp, new_fp.resolve())
            # Note : else: leave absolute paths

        record.data["file"] = str(new_fp)

    # Note : no named arguments (multiprocessing)
    def get_pdf(self, item: dict) -> dict:
        """Get PDFs (based on the package endpoints in the settings)"""

        record_dict = item["record"]

        if str(colrev.record.RecordState.rev_prescreen_included) != str(
            record_dict["colrev_status"]
        ):
            if "file" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.remove_field(key="file")
                return record.get_data()
            return record_dict

        record = colrev.record.Record(data=record_dict)

        for (
            pdf_get_package_endpoint
        ) in self.review_manager.settings.pdf_get.pdf_get_package_endpoints:
            endpoint_dict = self.package_manager.load_packages(
                package_type=colrev.env.package_manager.PackageEndpointType.pdf_get,
                selected_packages=[pdf_get_package_endpoint],
                operation=self,
                only_ci_supported=self.review_manager.in_ci_environment(),
            )
            if pdf_get_package_endpoint["endpoint"] not in endpoint_dict:
                self.review_manager.logger.info(
                    f'Skip {pdf_get_package_endpoint["endpoint"]} (not available)'
                )
                continue

            endpoint = endpoint_dict[pdf_get_package_endpoint["endpoint"]]
            endpoint.get_pdf(self, record)  # type: ignore

            if "file" in record.data:
                self.review_manager.report_logger.info(
                    f"{endpoint.settings.endpoint}"  # type: ignore
                    f'({record_dict["ID"]}): retrieved .../{Path(record_dict["file"]).name}'
                )
                if (
                    colrev.record.RecordState.rev_prescreen_included
                    == record.data["colrev_status"]
                ):
                    record.data.update(
                        colrev_status=colrev.record.RecordState.pdf_imported
                    )
                    self.review_manager.logger.info(
                        f" {colors.GREEN}{record.data['ID']}".ljust(46)
                        + f"rev_prescreen_included → pdf_imported{colors.END}"
                    )
                return record.get_data()

        if self.review_manager.settings.pdf_get.pdf_required_for_screen_and_synthesis:
            self.review_manager.logger.info(
                f" {colors.ORANGE}{record.data['ID']}".ljust(46)
                + f"rev_prescreen_included → pdf_needs_manual_retrieval{colors.END}"
            )

            record.data.update(
                colrev_status=colrev.record.RecordState.pdf_needs_manual_retrieval
            )
        else:
            self.review_manager.logger.info(
                f" {colors.ORANGE}{record.data['ID']}".ljust(46)
                + f"rev_prescreen_included → pdf_prepared{colors.END}"
            )

            record.data.update(colrev_status=colrev.record.RecordState.pdf_prepared)

        return record.get_data()

    def __relink_pdfs(
        self,
        *,
        records: typing.Dict[str, typing.Dict],
    ) -> typing.Dict[str, typing.Dict]:
        # pylint: disable=too-many-branches

        # Relink files in source file
        corresponding_origin: str
        source_records: typing.List[typing.Dict] = []
        for source in self.review_manager.settings.sources:
            if source.endpoint != "colrev.pdfs_dir":
                continue

            if not source.filename.is_file():
                continue

            corresponding_origin = str(source.filename)
            with open(source.filename, encoding="utf8") as target_db:
                source_records_dict = self.review_manager.dataset.load_records_dict(
                    load_str=target_db.read()
                )
            source_records = list(source_records_dict.values())

            self.review_manager.logger.info("Calculate colrev_pdf_ids")
            pdf_candidates = {
                pdf_candidate.relative_to(
                    self.review_manager.path
                ): colrev.record.Record.get_colrev_pdf_id(pdf_path=pdf_candidate)
                for pdf_candidate in list(self.review_manager.pdf_dir.glob("**/*.pdf"))
            }

            for record in records.values():
                if "file" not in record:
                    continue

                # Note: we check the source_records based on the cpids
                # in the record because cpids are not stored in the source_record
                # (pdf hashes may change after import/preparation)
                source_rec = {}
                if corresponding_origin != "":
                    source_origin_l = [
                        o for o in record["colrev_origin"] if corresponding_origin in o
                    ]
                    if len(source_origin_l) == 1:
                        source_origin = source_origin_l[0]
                        source_origin = source_origin.replace(
                            f"{corresponding_origin}/", ""
                        )
                        source_rec_l = [
                            s for s in source_records if s["ID"] == source_origin
                        ]
                        if len(source_rec_l) == 1:
                            source_rec = source_rec_l[0]

                if source_rec:
                    if (self.review_manager.path / Path(record["file"])).is_file() and (
                        self.review_manager.path / Path(source_rec["file"])
                    ).is_file():
                        continue
                else:
                    if (self.review_manager.path / Path(record["file"])).is_file():
                        continue

                self.review_manager.logger.info(record["ID"])

                for pdf_candidate, cpid in pdf_candidates.items():
                    if record.get("colrev_pdf_id", "") == cpid:
                        record["file"] = str(pdf_candidate)
                        source_rec["file"] = str(pdf_candidate)

                        self.review_manager.logger.info(
                            f"Found and linked PDF: {pdf_candidate}"
                        )
                        break

            if len(source_records) > 0:
                source_records_dict = {r["ID"]: r for r in source_records}
                self.review_manager.dataset.save_records_dict_to_file(
                    records=source_records_dict, save_path=source.filename
                )

            self.review_manager.dataset.add_changes(path=source.filename)

        return records

    def relink_pdfs(self) -> None:
        """Relink record files to the corresponding PDFs (if available)"""

        self.review_manager.logger.info(
            "Checking PDFs in same directory to reassig when the cpid is identical"
        )
        records = self.review_manager.dataset.load_records_dict()
        records = self.__relink_pdfs(records=records)

        self.review_manager.dataset.save_records_dict(records=records)

        self.review_manager.dataset.add_record_changes()
        self.review_manager.create_commit(msg="Relink PDFs")

    def check_existing_unlinked_pdfs(
        self,
        *,
        records: dict,
    ) -> dict:
        """Check for PDFs that are in the pdfs directory but not linked in the record file"""

        linked_pdfs = [
            str(Path(x["file"]).resolve()) for x in records.values() if "file" in x
        ]

        pdf_files = glob(str(self.review_manager.pdf_dir) + "/**.pdf", recursive=True)
        unlinked_pdfs = [
            Path(x)
            for x in pdf_files
            if str(Path(x).resolve()) not in linked_pdfs
            and not any(kw in x for kw in ["_wo_lp.pdf", "_wo_cp.pdf", "_ocr.pdf"])
        ]

        if len(unlinked_pdfs) == 0:
            return records

        grobid_service = self.review_manager.get_grobid_service()
        grobid_service.start()
        self.review_manager.logger.info("Check unlinked PDFs")
        for file in unlinked_pdfs:
            msg = f"Check unlinked PDF: {file.relative_to(self.review_manager.path)}"
            self.review_manager.logger.info(msg)
            if file.stem not in records.keys():
                tei = self.review_manager.get_tei(pdf_path=file)
                pdf_record = tei.get_metadata()

                if "error" in pdf_record:
                    continue

                max_similarity = 0.0
                max_sim_record = None
                for record in records.values():
                    sim = colrev.record.Record.get_record_similarity(
                        record_a=colrev.record.Record(data=pdf_record),
                        record_b=colrev.record.Record(data=record.copy()),
                    )
                    if sim > max_similarity:
                        max_similarity = sim
                        max_sim_record = record
                if max_sim_record:
                    if max_similarity > 0.5:
                        if (
                            colrev.record.RecordState.pdf_prepared
                            == max_sim_record["colrev_status"]
                        ):
                            continue

                        record = colrev.record.Record(data=max_sim_record)
                        record.update_field(
                            key="file",
                            value=str(file),
                            source="linking-available-files",
                        )
                        self.import_pdf(record=record)
                        if (
                            colrev.record.RecordState.rev_prescreen_included
                            == record.data["colrev_status"]
                        ):
                            record.set_status(
                                target_state=colrev.record.RecordState.pdf_imported
                            )

                        self.review_manager.report_logger.info(
                            "linked unlinked pdf:" f" {file.name}"
                        )
                        self.review_manager.logger.info(
                            "linked unlinked pdf:" f" {file.name}"
                        )
                        # max_sim_record = \
                        #     pdf_prep.validate_pdf_metadata(max_sim_record)
                        # colrev_status = max_sim_record['colrev_status']
                        # if RecordState.pdf_needs_manual_preparation == colrev_status:
                        #     # revert?
            else:
                record = records[file.stem]
                self.link_pdf(record=colrev.record.Record(data=record))

        self.review_manager.dataset.save_records_dict(records=records)

        return records

    def __rename_pdf(
        self,
        *,
        record_dict: dict,
        file: Path,
        new_filename: Path,
        pdfs_search_file: Path,
    ) -> None:
        record_dict["file"] = new_filename

        if "colrev_masterdata_provenance" in record_dict:
            for value in record_dict["colrev_masterdata_provenance"].values():
                if str(file) == value.get("source", ""):
                    value["source"] = str(new_filename)

        if "data_provenance" in record_dict:
            for value in record_dict["data_provenance"].values():
                if str(file) == value.get("source", ""):
                    value["source"] = str(new_filename)

        if pdfs_search_file.is_file():
            colrev.env.utils.inplace_change(
                filename=pdfs_search_file,
                old_string="{" + str(file) + "}",
                new_string="{" + str(new_filename) + "}",
            )

        if not file.is_file():
            corrected_path = Path(str(file).replace("  ", " "))
            if corrected_path.is_file():
                file = corrected_path

        if file.is_file():
            file.rename(new_filename)
        elif file.is_symlink():
            os.rename(str(file), str(new_filename))

        record_dict["file"] = str(new_filename)
        self.review_manager.logger.info(f"rename {file.name} > {new_filename}")
        if (
            colrev.record.RecordState.rev_prescreen_included
            == record_dict["colrev_status"]
        ):
            record = colrev.record.Record(data=record_dict)
            record.set_status(target_state=colrev.record.RecordState.pdf_imported)

    def rename_pdfs(self) -> None:
        """Rename the PDFs"""

        self.review_manager.logger.info("Rename PDFs")

        records = self.review_manager.dataset.load_records_dict()

        # We may use other pdfs_search_files from the sources:
        # review_manager.settings.sources
        pdfs_search_file = Path("data/search/pdfs.bib")

        for record_dict in records.values():
            if "file" not in record_dict:
                continue
            if record_dict[
                "colrev_status"
            ] not in colrev.record.RecordState.get_post_x_states(
                state=colrev.record.RecordState.md_processed
            ):
                continue

            file = Path(record_dict["file"])
            new_filename = file.parents[0] / Path(f"{record_dict['ID']}{file.suffix}")
            # Possible option: move to top (pdfs) directory:
            # new_filename = self.review_manager.PDF_DIR_RELATIVE / Path(
            #     f"{record['ID']}.pdf"
            # )
            if str(file) == str(new_filename):
                continue

            self.__rename_pdf(
                record_dict=record_dict,
                file=file,
                new_filename=new_filename,
                pdfs_search_file=pdfs_search_file,
            )

        self.review_manager.dataset.save_records_dict(records=records)

        if pdfs_search_file.is_file():
            self.review_manager.dataset.add_changes(path=pdfs_search_file)
        self.review_manager.dataset.add_record_changes()

    def __get_data(self) -> dict:
        # pylint: disable=duplicate-code

        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())

        nr_tasks = len(
            [
                x
                for x in record_header_list
                if colrev.record.RecordState.rev_prescreen_included
                == x["colrev_status"]
            ]
        )

        items = self.review_manager.dataset.read_next_record(
            conditions=[
                {"colrev_status": colrev.record.RecordState.rev_prescreen_included}
            ],
        )

        self.to_retrieve = nr_tasks

        pdf_get_data = {
            "nr_tasks": nr_tasks,
            "items": [{"record": item} for item in items],
        }

        return pdf_get_data

    def _print_stats(self, *, retrieved_record_list: list) -> None:
        self.retrieved = len([r for r in retrieved_record_list if "file" in r])

        self.not_retrieved = self.to_retrieve - self.retrieved

        retrieved_string = "Overall pdf_imported".ljust(34)
        if self.retrieved == 0:
            retrieved_string += f"{self.retrieved}".rjust(6, " ")
            retrieved_string += " PDFs"
        elif self.retrieved == 1:
            retrieved_string += f"{colors.GREEN}"
            retrieved_string += f"{self.retrieved}".rjust(6, " ")
            retrieved_string += f"{colors.END} PDF"
        else:
            retrieved_string += f"{colors.GREEN}"
            retrieved_string += f"{self.retrieved}".rjust(6, " ")
            retrieved_string += f"{colors.END} PDFs"

        not_retrieved_string = "Overall pdf_needs_manual_retrieval".ljust(34)
        if self.not_retrieved == 0:
            not_retrieved_string += f"{self.not_retrieved}".rjust(6, " ")
            not_retrieved_string += " PDFs"
        elif self.not_retrieved == 1:
            not_retrieved_string += f"{colors.ORANGE}"
            not_retrieved_string += f"{self.not_retrieved}".rjust(6, " ")
            not_retrieved_string += f"{colors.END} PDF"
        else:
            not_retrieved_string += f"{colors.ORANGE}"
            not_retrieved_string += f"{self.not_retrieved}".rjust(6, " ")
            not_retrieved_string += f"{colors.END} PDFs"

        self.review_manager.logger.info(retrieved_string)
        self.review_manager.logger.info(not_retrieved_string)

    def __set_status_if_pdf_linked(self, *, records: dict) -> dict:
        for record_dict in records.values():
            if record_dict["colrev_status"] in [
                colrev.record.RecordState.rev_prescreen_included,
                colrev.record.RecordState.pdf_needs_manual_retrieval,
            ]:
                record = colrev.record.Record(data=record_dict)
                if "file" in record_dict:
                    if any(
                        Path(fpath).is_file()
                        for fpath in record.data["file"].split(";")
                    ):
                        if (
                            colrev.record.RecordState.rev_prescreen_included
                            == record.data["colrev_status"]
                        ):
                            record.set_status(
                                target_state=colrev.record.RecordState.pdf_imported
                            )
                    else:
                        self.review_manager.logger.warning(
                            "Remove non-existent file link "
                            f'({record_dict["ID"]}: {record_dict["file"]}'
                        )
                        record.remove_field(key="file")
        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()

        return records

    def setup_custom_script(self) -> None:
        """Setup a custom pfd-get script"""

        filedata = colrev.env.utils.get_package_file_content(
            file_path=Path("template/custom_scripts/custom_pdf_get_script.py")
        )
        if filedata:
            with open("custom_pdf_get_script.py", "w", encoding="utf-8") as file:
                file.write(filedata.decode("utf-8"))

        self.review_manager.dataset.add_changes(path=Path("custom_pdf_get_script.py"))

        self.review_manager.settings.pdf_get.pdf_get_man_package_endpoints.append(
            {"endpoint": "custom_pdf_get_script"}
        )

        self.review_manager.save_settings()

    @colrev.operation.Operation.decorate()
    def main(self) -> None:
        """Get PDFs (main entrypoint)"""

        if self.review_manager.in_ci_environment():
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="colrev pdf-prep",
                detailed_trace="pdf-prep not available in ci environment",
            )

        self.review_manager.logger.info("Get PDFs")
        self.review_manager.logger.info(
            "Get PDFs of prescreen-included records from local and remote sources."
        )
        self.review_manager.logger.info(
            "PDFs are stored in the directory data/pdfs "
            f"({colors.ORANGE}colrev pdfs --dir{colors.END})"
        )
        self.review_manager.logger.info(
            "See https://colrev.readthedocs.io/en/latest/manual/pdf_retrieval/pdf_get.html"
        )

        records = self.review_manager.dataset.load_records_dict()
        records = self.__set_status_if_pdf_linked(records=records)
        records = self.check_existing_unlinked_pdfs(records=records)

        pdf_get_data = self.__get_data()

        if pdf_get_data["nr_tasks"] == 0:
            self.review_manager.logger.info("No additional pdfs to retrieve")
        else:
            self.review_manager.logger.info(
                "PDFs to get".ljust(38) + f'{pdf_get_data["nr_tasks"]} PDFs'
            )

            pool = Pool(4)
            retrieved_record_list = pool.map(self.get_pdf, pdf_get_data["items"])
            pool.close()
            pool.join()

            self.review_manager.dataset.save_records_dict(
                records={r["ID"]: r for r in retrieved_record_list}, partial=True
            )

            self._print_stats(retrieved_record_list=retrieved_record_list)

        # Note: rename should be after copy.
        # Note : do not pass records as an argument.
        if self.review_manager.settings.pdf_get.rename_pdfs:
            self.rename_pdfs()

        self.review_manager.create_commit(msg="Get PDFs")
        self.review_manager.logger.info(
            f"{colors.GREEN}Completed pdf-get operation{colors.END}"
        )



#! /usr/bin/env python
"""Checker for identical-values-between-title-and-container."""
from __future__ import annotations

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class IdenticalValuesChecker:
    """The IdenticalValuesChecker"""

    msg = "identical-values-between-title-and-container"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the identical-values-between-title-and-container checks"""

        if self.__identical_values_between_title_and_container(record=record):
            record.add_masterdata_provenance_note(key="title", note=self.msg)
        else:
            record.remove_masterdata_provenance_note(key="title", note=self.msg)

    def __identical_values_between_title_and_container(
        self, *, record: colrev.record.Record
    ) -> bool:
        if record.data.get("title", "UNKNOWN") == "UNKNOWN":
            return False
        if (
            "booktitle" in record.data
            and "title" in record.data
            and record.data["title"].lower().replace("the ", "")
            == record.data["booktitle"].lower().replace("the ", "")
        ):
            return True
        if (
            "journal" in record.data
            and "title" in record.data
            and record.data["title"].lower().replace("the ", "")
            == record.data["journal"].lower().replace("the ", "")
        ):
            return True
        return False


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(IdenticalValuesChecker(quality_model))



#! /usr/bin/env python
"""Checker for name particles."""
from __future__ import annotations

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class NameParticlesChecker:
    """The NameParticlesChecker"""

    fields_to_check = ["author", "editor"]

    msg = "name-particles"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the name-particles checks"""

        for key in self.fields_to_check:
            if key not in record.data:
                continue

            names = record.data[key].split(" and ")
            if any(self.__particle_defect(name=name) for name in names):
                record.add_masterdata_provenance_note(key=key, note=self.msg)
            else:
                record.remove_masterdata_provenance_note(key=key, note=self.msg)

    def __particle_defect(self, *, name: str) -> bool:
        if name.endswith(" vom") or name.startswith("vom "):
            return True

        if name.endswith(" von") or name.startswith("von "):
            return True

        return False


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(NameParticlesChecker(quality_model))



#! /usr/bin/env python
"""Checker for page-range."""
from __future__ import annotations

import re

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class PageRangeChecker:
    """The PageRangeChecker"""

    msg = "page-range"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the page-range checks"""

        if "pages" not in record.data:
            return
        if not re.match(r"^\d+\-\-\d+$", record.data["pages"]):
            return

        if self.__pages_descending(pages=record.data["pages"]):
            record.add_masterdata_provenance_note(key="pages", note=self.msg)
        else:
            record.remove_masterdata_provenance_note(key="pages", note=self.msg)

    def __pages_descending(self, *, pages: str) -> bool:
        from_page, to_page = re.findall(r"(\d+)", pages)
        if int(from_page) > int(to_page):
            return True

        return False


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(PageRangeChecker(quality_model))



#! /usr/bin/env python
"""Prescreen based on a table"""
from __future__ import annotations

import csv
import typing
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.record
import colrev.ui_cli.cli_colors as colors

if typing.TYPE_CHECKING:
    import colrev.ops.prescreen.Prescreen

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.PrescreenPackageEndpointInterface
)
@dataclass
class TablePrescreen(JsonSchemaMixin):

    """Table-based prescreen (exported and imported)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False
    export_todos_only: bool = True

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def export_table(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        records: dict,
        split: list,
        export_table_format: str = "csv",
    ) -> None:
        """Export a prescreen table"""

        # gh_issue https://github.com/CoLRev-Environment/colrev/issues/73
        # add delta (records not yet in the table)
        # instead of overwriting
        # export_table_format as a settings parameter

        prescreen_operation.review_manager.logger.info("Loading records for export")

        tbl = []
        for record in records.values():
            if record["colrev_status"] not in [
                colrev.record.RecordState.md_processed,
                colrev.record.RecordState.rev_prescreen_excluded,
                colrev.record.RecordState.rev_prescreen_included,
                colrev.record.RecordState.pdf_needs_manual_retrieval,
                colrev.record.RecordState.pdf_imported,
                colrev.record.RecordState.pdf_not_available,
                colrev.record.RecordState.pdf_needs_manual_preparation,
                colrev.record.RecordState.pdf_prepared,
                colrev.record.RecordState.rev_excluded,
                colrev.record.RecordState.rev_included,
                colrev.record.RecordState.rev_synthesized,
            ]:
                continue

            if len(split) > 0:
                if record["ID"] not in split:
                    continue

            if colrev.record.RecordState.md_processed == record["colrev_status"]:
                inclusion_1 = "TODO"
            elif self.export_todos_only:
                continue
            elif (
                colrev.record.RecordState.rev_prescreen_excluded
                == record["colrev_status"]
            ):
                inclusion_1 = "out"
            else:
                inclusion_1 = "in"

            # pylint: disable=duplicate-code
            row = {
                "ID": record["ID"],
                "author": record.get("author", ""),
                "title": record.get("title", ""),
                "journal": record.get("journal", ""),
                "booktitle": record.get("booktitle", ""),
                "year": record.get("year", ""),
                "volume": record.get("volume", ""),
                "number": record.get("number", ""),
                "pages": record.get("pages", ""),
                "doi": record.get("doi", ""),
                "abstract": record.get("abstract", ""),
                "presceen_inclusion": inclusion_1,
            }
            tbl.append(row)

        if export_table_format.lower() == "csv":
            screen_df = pd.DataFrame(tbl)
            screen_df.to_csv("prescreen.csv", index=False, quoting=csv.QUOTE_ALL)
            prescreen_operation.review_manager.logger.info("Created prescreen.csv")

        if export_table_format.lower() == "xlsx":
            screen_df = pd.DataFrame(tbl)
            screen_df.to_excel("prescreen.xlsx", index=False, sheet_name="screen")
            prescreen_operation.review_manager.logger.info("Created prescreen.xlsx")

        prescreen_operation.review_manager.logger.info(
            f"To prescreen records, {colors.ORANGE}enter [in|out] "
            f"in the presceen_inclusion column.{colors.END}"
        )
        prescreen_operation.review_manager.logger.info(
            f"Afterwards, run {colors.ORANGE}colrev prescreen --import_table "
            f"prescreen.{export_table_format.lower()}{colors.END}"
        )

    def import_table(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        records: dict,
        import_table_path: str = "prescreen.csv",
    ) -> None:
        """Import a prescreen table"""

        # pylint: disable=too-many-branches

        prescreen_operation.review_manager.logger.info(f"Load {import_table_path}")

        # pylint: disable=duplicate-code
        if not Path(import_table_path).is_file():
            prescreen_operation.review_manager.logger.error(
                f"Did not find {import_table_path} - exiting."
            )
            return
        prescreen_df = pd.read_csv(import_table_path)
        prescreen_df.fillna("", inplace=True)
        prescreened_records = prescreen_df.to_dict("records")

        if "presceen_inclusion" not in prescreened_records[0]:
            prescreen_operation.review_manager.logger.warning(
                "presceen_inclusion column missing"
            )
            return

        prescreen_included = 0
        prescreen_excluded = 0
        nr_todo = 0
        prescreen_operation.review_manager.logger.info("Update prescreen results")
        for prescreened_record in prescreened_records:
            if prescreened_record.get("ID", "") in records:
                record = colrev.record.Record(
                    data=records[prescreened_record.get("ID", "")]
                )
                if record.data[
                    "colrev_status"
                ] in colrev.record.RecordState.get_post_x_states(
                    state=colrev.record.RecordState.rev_prescreen_included
                ):
                    if (
                        "in" == prescreened_record.data.get("presceen_inclusion", "")
                        and colrev.record.RecordState.rev_prescreen_excluded
                        != record.data["colrev_status"]
                    ):
                        continue

                if prescreened_record.get("presceen_inclusion", "") == "out":
                    if (
                        record.data["colrev_status"]
                        != colrev.record.RecordState.rev_prescreen_excluded
                    ):
                        prescreen_excluded += 1
                    record.set_status(
                        target_state=colrev.record.RecordState.rev_prescreen_excluded
                    )

                elif prescreened_record.get("presceen_inclusion", "") == "in":
                    if (
                        record.data["colrev_status"]
                        != colrev.record.RecordState.rev_prescreen_included
                    ):
                        prescreen_included += 1
                    record.set_status(
                        target_state=colrev.record.RecordState.rev_prescreen_included
                    )
                elif prescreened_record.get("presceen_inclusion", "") == "TODO":
                    nr_todo += 1
                else:
                    prescreen_operation.review_manager.logger.warning(
                        "Invalid value in prescreen_inclusion: "
                        f"{prescreened_record.get('presceen_inclusion', '')} "
                        f"({prescreened_record.get('ID', 'NO_ID')})"
                    )

            else:
                prescreen_operation.review_manager.logger.warning(
                    f"ID not in records: {prescreened_record.get('ID', '')}"
                )

        prescreen_operation.review_manager.logger.info(
            f" {colors.GREEN}{prescreen_included} records prescreen_included{colors.END}"
        )
        prescreen_operation.review_manager.logger.info(
            f" {colors.RED}{prescreen_excluded} records prescreen_excluded{colors.END}"
        )

        prescreen_operation.review_manager.logger.info(
            f" {colors.ORANGE}{nr_todo} records to prescreen{colors.END}"
        )

        prescreen_operation.review_manager.dataset.save_records_dict(records=records)
        prescreen_operation.review_manager.dataset.add_record_changes()

        prescreen_operation.review_manager.logger.info("Completed import")

    def run_prescreen(
        self,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        records: dict,
        split: list,
    ) -> dict:
        """Prescreen records based on screening tables"""

        if input("create prescreen table [y,n]?") == "y":
            self.export_table(
                prescreen_operation=prescreen_operation, records=records, split=split
            )

        if input("import prescreen table [y,n]?") == "y":
            self.import_table(prescreen_operation=prescreen_operation, records=records)

        if prescreen_operation.review_manager.dataset.has_changes():
            if input("create commit [y,n]?") == "y":
                prescreen_operation.review_manager.create_commit(
                    msg="Pre-screen (table)",
                    manual_author=True,
                )
        return records



#! /usr/bin/env python
"""SearchSource: Pubmed"""
from __future__ import annotations

import typing
from copy import deepcopy
from dataclasses import dataclass
from multiprocessing import Lock
from pathlib import Path
from sqlite3 import OperationalError
from typing import Optional
from urllib.parse import urlparse
from xml.etree import ElementTree  # nosec
from xml.etree.ElementTree import Element  # nosec

import defusedxml
import requests
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin
from defusedxml.lxml import fromstring
from lxml.etree import XMLSyntaxError

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.load_utils_table
import colrev.ops.search
import colrev.record

defusedxml.defuse_stdlib()


# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class PubMedSearchSource(JsonSchemaMixin):
    """SearchSource for Pubmed"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "pubmedid"
    search_type = colrev.settings.SearchType.DB
    endpoint = "colrev.pubmed"
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "PubMed"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/pubmed.md"
    )
    __pubmed_md_filename = Path("data/search/md_pubmed.bib")

    def __init__(
        self,
        *,
        source_operation: colrev.operation.Operation,
        settings: Optional[dict] = None,
    ) -> None:
        if settings:
            # Pubmed as a search_source
            self.search_source = from_dict(
                data_class=self.settings_class, data=settings
            )
        else:
            # Pubmed as an md-prep source
            pubmed_md_source_l = [
                s
                for s in source_operation.review_manager.settings.sources
                if s.filename == self.__pubmed_md_filename
            ]
            if pubmed_md_source_l:
                self.search_source = pubmed_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint=self.endpoint,
                    filename=self.__pubmed_md_filename,
                    search_type=colrev.settings.SearchType.OTHER,
                    search_parameters={},
                    comment="",
                )

            self.pubmed_lock = Lock()

        self.review_manager = source_operation.review_manager
        self.quality_model = self.review_manager.get_qm()
        _, self.email = source_operation.review_manager.get_committer()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Pubmed"""

        result = {"confidence": 0.0}

        # Simple heuristic:
        if "PMID,Title,Authors,Citation,First Author,Journal/Book," in data:
            result["confidence"] = 1.0
            return result
        if "PMID- " in data:
            result["confidence"] = 0.7
            return result

        if "pmid " in data:
            if data.count(" pmid ") > data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        if params is None:
            operation.add_interactively(endpoint=cls.endpoint)
            return

        host = urlparse(params).hostname

        if host and host.endswith("pubmed.ncbi.nlm.nih.gov"):
            params = params.replace("https://pubmed.ncbi.nlm.nih.gov/?term=", "")

            filename = operation.get_unique_filename(
                file_path_string=f"pubmed_{params.replace('&sort=', '')}"
            )
            params = (
                "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=pubmed&term="
                + params
            )
            add_source = colrev.settings.SearchSource(
                endpoint=cls.endpoint,
                filename=filename,
                search_type=colrev.settings.SearchType.DB,
                search_parameters={"query": params},
                comment="",
            )
            operation.review_manager.settings.sources.append(add_source)
            return

        raise NotImplementedError

    def __validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        # TO DO
        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if source.filename.name != self.__pubmed_md_filename.name:
            if "query" not in source.search_parameters:
                raise colrev_exceptions.InvalidQueryException(
                    f"Source missing query search_parameter ({source.filename})"
                )

            # if "query_file" in source.search_parameters:
            # ...

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def check_availability(
        self, *, source_operation: colrev.operation.Operation
    ) -> None:
        """Check status (availability) of the Pubmed API"""

        try:
            # pylint: disable=duplicate-code
            test_rec = {
                "author": "Nazzaro, P and Manzari, M and Merlo, M and Triggiani, R and "
                "Scarano, A and Ciancio, L and Pirrelli, A",
                "title": "Distinct and combined vascular effects of ACE blockade and "
                "HMG-CoA reductase inhibition in hypertensive subjects",
                "ENTRYTYPE": "article",
                "pubmedid": "10024335",
            }
            returned_record_dict = self.__pubmed_query_id(
                pubmed_id=test_rec["pubmedid"],
                timeout=20,
            )

            if returned_record_dict:
                assert returned_record_dict["title"] == test_rec["title"]
                assert returned_record_dict["author"] == test_rec["author"]
            else:
                if not source_operation.force_mode:
                    raise colrev_exceptions.ServiceNotAvailableException("Pubmed")
        except (requests.exceptions.RequestException, IndexError) as exc:
            print(exc)
            if not source_operation.force_mode:
                raise colrev_exceptions.ServiceNotAvailableException("Pubmed") from exc

    @classmethod
    def __get_author_string_from_node(cls, *, author_node: Element) -> str:
        authors_string = ""
        author_last_name_node = author_node.find("LastName")
        if author_last_name_node is not None:
            if author_last_name_node.text is not None:
                authors_string += author_last_name_node.text
        author_fore_name_node = author_node.find("ForeName")
        if author_fore_name_node is not None:
            if author_fore_name_node.text is not None:
                authors_string += ", "
                authors_string += author_fore_name_node.text
        return authors_string

    @classmethod
    def __get_author_string(cls, *, root) -> str:  # type: ignore
        authors_list = []
        for author_node in root.xpath(
            "/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/AuthorList/Author"
        ):
            authors_list.append(
                cls.__get_author_string_from_node(author_node=author_node)
            )
        return " and ".join(authors_list)

    @classmethod
    def __get_title_string(cls, *, root) -> str:  # type: ignore
        title = root.xpath(
            "/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/ArticleTitle"
        )
        if title:
            if title[0].text:
                title = title[0].text.strip().rstrip(".")
                if title.startswith("[") and title.endswith("]"):
                    title = title[1:-1]
                return title
        return ""

    @classmethod
    def __get_abstract_string(cls, *, root) -> str:  # type: ignore
        abstract = root.xpath(
            "/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/Abstract"
        )
        if abstract:
            return ElementTree.tostring(abstract[0], encoding="unicode")
        return ""

    @classmethod
    def __pubmed_xml_to_record(cls, *, root) -> dict:  # type: ignore
        retrieved_record_dict: dict = {"ENTRYTYPE": "misc"}

        pubmed_article = root.find("PubmedArticle")
        if pubmed_article is None:
            return {}
        if pubmed_article.find("MedlineCitation") is None:
            return {}

        retrieved_record_dict.update(title=cls.__get_title_string(root=root))
        retrieved_record_dict.update(author=cls.__get_author_string(root=root))

        journal_path = "/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/Journal"
        journal_name = root.xpath(journal_path + "/ISOAbbreviation")
        if journal_name:
            retrieved_record_dict.update(ENTRYTYPE="article")
            retrieved_record_dict.update(journal=journal_name[0].text)

        volume = root.xpath(journal_path + "/JournalIssue/Volume")
        if volume:
            retrieved_record_dict.update(volume=volume[0].text)

        number = root.xpath(journal_path + "/JournalIssue/Issue")
        if number:
            retrieved_record_dict.update(number=number[0].text)

        year = root.xpath(journal_path + "/JournalIssue/PubDate/Year")
        if year:
            retrieved_record_dict.update(year=year[0].text)

        retrieved_record_dict.update(abstract=cls.__get_abstract_string(root=root))

        article_id_list = root.xpath(
            "/PubmedArticleSet/PubmedArticle/PubmedData/ArticleIdList"
        )
        for article_id in article_id_list[0]:
            id_type = article_id.attrib.get("IdType")
            if article_id.attrib.get("IdType") == "pubmed":
                retrieved_record_dict.update(pubmedid=article_id.text.upper())
            elif article_id.attrib.get("IdType") == "doi":
                retrieved_record_dict.update(doi=article_id.text.upper())
            else:
                retrieved_record_dict[id_type] = article_id.text

        retrieved_record_dict = {
            k: v for k, v in retrieved_record_dict.items() if v != ""
        }
        if (
            retrieved_record_dict.get("pii", "pii").lower()
            == retrieved_record_dict.get("doi", "doi").lower()
        ):
            del retrieved_record_dict["pii"]

        return retrieved_record_dict

    def __get_pubmed_ids(self, query: str, retstart: int) -> typing.List[str]:
        headers = {"user-agent": f"{__name__} (mailto:{self.email})"}
        session = self.review_manager.get_cached_session()
        if not query.startswith("https://pubmed.ncbi.nlm.nih.gov/?term="):
            query = "https://pubmed.ncbi.nlm.nih.gov/?term=" + query
        ret = session.request(
            "GET", query + f"&retstart={retstart}", headers=headers, timeout=30
        )

        ret.raise_for_status()
        if ret.status_code != 200:
            # review_manager.logger.debug(
            #     f"crossref_query failed with status {ret.status_code}"
            # )
            return []

        root = fromstring(str.encode(ret.text))
        return [
            x.text
            for x_el in root.findall("IdList")
            for x in x_el
            if x.text is not None
        ]

    def __pubmed_query_id(
        self,
        *,
        pubmed_id: str,
        timeout: int = 60,
    ) -> dict:
        """Retrieve records from Pubmed based on a query"""

        try:
            database = "pubmed"
            url = (
                "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?"
                + f"db={database}&id={pubmed_id}&rettype=xml&retmode=text"
            )

            headers = {"user-agent": f"{__name__} (mailto:{self.email})"}
            session = self.review_manager.get_cached_session()

            # review_manager.logger.debug(url)
            ret = session.request("GET", url, headers=headers, timeout=timeout)
            ret.raise_for_status()
            if ret.status_code != 200:
                # review_manager.logger.debug(
                #     f"crossref_query failed with status {ret.status_code}"
                # )
                return {"pubmed_id": pubmed_id}

            root = fromstring(str.encode(ret.text))
            retrieved_record = self.__pubmed_xml_to_record(root=root)
            if not retrieved_record:
                return {"pubmed_id": pubmed_id}
        except requests.exceptions.RequestException:
            return {"pubmed_id": pubmed_id}
        except XMLSyntaxError as exc:
            raise colrev_exceptions.RecordNotParsableException(
                "Error parsing xml"
            ) from exc
        # pylint: disable=duplicate-code
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                "sqlite, required for requests CachedSession "
                "(possibly caused by concurrent operations)"
            ) from exc

        return retrieved_record

    def __get_masterdata_record(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool,
        timeout: int,
    ) -> colrev.record.Record:
        try:
            retrieved_record_dict = self.__pubmed_query_id(
                pubmed_id=record.data["pubmedid"],
                timeout=timeout,
            )

            retries = 0
            while (
                not retrieved_record_dict
                and retries < prep_operation.max_retries_on_error
            ):
                retries += 1

                retrieved_record_dict = self.__pubmed_query_id(
                    pubmed_id=record.data["pubmedid"],
                    timeout=timeout,
                )

            if not retrieved_record_dict:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="Pubmed: no records retrieved"
                )

            retrieved_record = colrev.record.Record(data=retrieved_record_dict)

            similarity = colrev.record.PrepRecord.get_retrieval_similarity(
                record_original=record, retrieved_record_original=retrieved_record
            )
            # prep_operation.review_manager.logger.debug("Found matching record")
            # prep_operation.review_manager.logger.debug(
            #     f"crossref similarity: {similarity} "
            #     f"(>{prep_operation.retrieval_similarity})"
            # )
            self.review_manager.logger.debug(
                f"pubmed similarity: {similarity} "
                f"(<{prep_operation.retrieval_similarity})"
            )

            try:
                self.pubmed_lock.acquire(timeout=60)

                # Note : need to reload file because the object is not shared between processes
                pubmed_feed = self.search_source.get_feed(
                    review_manager=self.review_manager,
                    source_identifier=self.source_identifier,
                    update_only=False,
                )

                pubmed_feed.set_id(record_dict=retrieved_record.data)
                pubmed_feed.add_record(record=retrieved_record)

                record.merge(
                    merging_record=retrieved_record,
                    default_source=retrieved_record.data["colrev_origin"][0],
                )

                record.set_masterdata_complete(
                    source=retrieved_record.data["colrev_origin"][0],
                    masterdata_repository=self.review_manager.settings.is_curated_repo(),
                )
                record.set_status(target_state=colrev.record.RecordState.md_prepared)
                if save_feed:
                    pubmed_feed.save_feed_file()
                try:
                    self.pubmed_lock.release()
                except ValueError:
                    pass

                return record
            except (
                colrev_exceptions.InvalidMerge,
                colrev_exceptions.NotFeedIdentifiableException,
            ):
                try:
                    self.pubmed_lock.release()
                except ValueError:
                    pass

                return record

        except (
            requests.exceptions.RequestException,
            OSError,
            IndexError,
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            colrev_exceptions.RecordNotParsableException,
        ):
            pass
        return record

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Retrieve masterdata from Pubmed based on similarity with the record provided"""

        # To test the metadata provided for a particular pubmed-id use:
        # https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=10075143&rettype=xml&retmode=text

        if (
            len(record.data.get("title", "")) < 35
            and "colrev.pubmed.pubmedid" not in record.data
        ):
            return record

        # at this point, we coujld validate metadata
        # if "pubmedid" not in record.data:
        #    record = self.__check_doi_masterdata(record=record)

        # remove the following if we match basd on similarity
        if "colrev.pubmed.pubmedid" not in record.data:
            return record

        record = self.__get_masterdata_record(
            prep_operation=prep_operation,
            record=record,
            timeout=timeout,
            save_feed=save_feed,
        )

        return record

    def __get_pubmed_query_return(self) -> typing.Iterator[dict]:
        params = self.search_source.search_parameters

        retstart = 0
        while True:
            pubmed_ids = self.__get_pubmed_ids(query=params["query"], retstart=retstart)
            if not pubmed_ids:
                break
            for pubmed_id in pubmed_ids:
                yield self.__pubmed_query_id(pubmed_id=pubmed_id)

            retstart += 20

    def __run_parameter_search(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        pubmed_feed: colrev.ops.search_feed.GeneralOriginFeed,
        rerun: bool,
    ) -> None:
        if rerun:
            search_operation.review_manager.logger.info(
                "Performing a search of the full history (may take time)"
            )

        records = search_operation.review_manager.dataset.load_records_dict()
        try:
            for record_dict in self.__get_pubmed_query_return():
                # Note : discard "empty" records
                if "" == record_dict.get("author", "") and "" == record_dict.get(
                    "title", ""
                ):
                    search_operation.review_manager.logger.warning(
                        f"Skipped record: {record_dict}"
                    )
                    continue
                try:
                    pubmed_feed.set_id(record_dict=record_dict)
                except colrev_exceptions.NotFeedIdentifiableException:
                    continue

                prev_record_dict_version = {}
                if record_dict["ID"] in pubmed_feed.feed_records:
                    prev_record_dict_version = deepcopy(
                        pubmed_feed.feed_records[record_dict["ID"]]
                    )

                prep_record = colrev.record.PrepRecord(data=record_dict)

                if "colrev_data_provenance" in prep_record.data:
                    del prep_record.data["colrev_data_provenance"]

                added = pubmed_feed.add_record(record=prep_record)

                if added:
                    search_operation.review_manager.logger.info(
                        " retrieve pubmed-id=" + prep_record.data["pubmedid"]
                    )
                else:
                    pubmed_feed.update_existing_record(
                        records=records,
                        record_dict=prep_record.data,
                        prev_record_dict_version=prev_record_dict_version,
                        source=self.search_source,
                        update_time_variant_fields=rerun,
                    )

                # Note : only retrieve/update the latest deposits (unless in rerun mode)
                if not added and not rerun:
                    # problem: some publishers don't necessarily
                    # deposit papers chronologically
                    break

            pubmed_feed.print_post_run_search_infos(records=records)
            pubmed_feed.save_feed_file()
            search_operation.review_manager.dataset.save_records_dict(records=records)
            search_operation.review_manager.dataset.add_record_changes()

        except requests.exceptions.JSONDecodeError as exc:
            # watch github issue:
            # https://github.com/fabiobatalha/crossrefapi/issues/46
            if "504 Gateway Time-out" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Crossref (check https://status.crossref.org/)"
                )
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Crossref (check https://status.crossref.org/) ({exc})"
            )

    def __run_md_search_update(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        pubmed_feed: colrev.ops.search_feed.GeneralOriginFeed,
    ) -> None:
        records = search_operation.review_manager.dataset.load_records_dict()

        for feed_record_dict in pubmed_feed.feed_records.values():
            feed_record = colrev.record.Record(data=feed_record_dict)

            try:
                retrieved_record = self.__pubmed_query_id(
                    pubmed_id=feed_record_dict["pubmedid"]
                )

                if retrieved_record["pubmedid"] != feed_record.data["pubmedid"]:
                    continue

                pubmed_feed.set_id(record_dict=retrieved_record)
            except (
                colrev_exceptions.RecordNotFoundInPrepSourceException,
                colrev_exceptions.NotFeedIdentifiableException,
            ):
                continue

            prev_record_dict_version = {}
            if retrieved_record["ID"] in pubmed_feed.feed_records:
                prev_record_dict_version = pubmed_feed.feed_records[
                    retrieved_record["ID"]
                ]

            pubmed_feed.add_record(record=colrev.record.Record(data=retrieved_record))

            pubmed_feed.update_existing_record(
                records=records,
                record_dict=retrieved_record,
                prev_record_dict_version=prev_record_dict_version,
                source=self.search_source,
                update_time_variant_fields=True,
            )

        pubmed_feed.save_feed_file()
        pubmed_feed.print_post_run_search_infos(records=records)
        search_operation.review_manager.dataset.save_records_dict(records=records)
        search_operation.review_manager.dataset.add_record_changes()

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of Pubmed"""

        self.__validate_source()

        pubmed_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        if self.search_source.is_md_source() or self.search_source.is_quasi_md_source():
            self.__run_md_search_update(
                search_operation=search_operation,
                pubmed_feed=pubmed_feed,
            )

        else:
            self.__run_parameter_search(
                search_operation=search_operation,
                pubmed_feed=pubmed_feed,
                rerun=rerun,
            )

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".csv":
            csv_loader = colrev.ops.load_utils_table.CSVLoader(
                load_operation=load_operation, source=self.search_source
            )
            records = csv_loader.load()
            self.__load_fixes(records=records)
            return records

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def __load_fixes(
        self,
        records: typing.Dict,
    ) -> None:
        """Load fixes for Pubmed"""

        for record in records.values():
            if "author" in record and record["author"].count(",") >= 1:
                author_list = record["author"].split(", ")
                for i, author_part in enumerate(author_list):
                    author_field_parts = author_part.split(" ")
                    author_list[i] = (
                        author_field_parts[0] + ", " + " ".join(author_field_parts[1:])
                    )

                record["author"] = " and ".join(author_list)
            if "first_author" in record:
                del record["first_author"]
            if record.get("journal", "") != "":
                record["ENTRYTYPE"] = "article"
            if record.get("pii", "pii").lower() == record.get("doi", "doi").lower():
                del record["pii"]
            if record.get("nihms_id", "") == "nan":
                del record["nihms_id"]
            if "citation" in record:
                details_part = record["citation"]
                details_part = details_part[details_part.find(";") + 1 :]
                details_part = details_part[: details_part.find(".")]
                if ":" in details_part:
                    record["pages"] = details_part[details_part.find(":") + 1 :]
                    details_part = details_part[: details_part.find(":")]
                if "(" in details_part:
                    record["number"] = details_part[details_part.find("(") + 1 : -1]
                    details_part = details_part[: details_part.find("(")]
                record["volume"] = details_part
                del record["citation"]
            if "journal/book" in record:
                record["journal"] = record.pop("journal/book")

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for Pubmed"""

        if "colrev.pubmed.first_author" in record.data:
            record.remove_field(key="colrev.pubmed.first_author")
        if record.data.get("author") == "UNKNOWN" and "authors" in record.data:
            record.remove_field(key="author")
            record.rename_field(key="authors", new_key="author")

        if record.data.get("year") == "UNKNOWN":
            record.remove_field(key="year")
            if "colrev.pubmed.publication_year" in record.data:
                record.rename_field(
                    key="colrev.pubmed.publication_year", new_key="year"
                )

        if "author" in record.data:
            record.data["author"] = colrev.record.PrepRecord.format_author_field(
                input_string=record.data["author"]
            )

        # TBD: how to distinguish other types?
        record.change_entrytype(new_entrytype="article", qm=self.quality_model)

        return record



#! /usr/bin/env python
"""CoLRev pdf_prep_man operation: Prepare PDF documents manually."""
from __future__ import annotations

from pathlib import Path

import pandas as pd
from PyPDF2 import PdfFileReader
from PyPDF2 import PdfFileWriter
from PyPDF2.errors import PdfReadError

import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record


class PDFPrepMan(colrev.operation.Operation):
    """Prepare PDFs manually"""

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.pdf_prep_man,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.verbose = True

    def discard(self) -> None:
        """Discard records whose PDFs need manual preparation (set to pdf_not_available)"""

        records = self.review_manager.dataset.load_records_dict()
        for record_dict in records.values():
            if (
                record_dict["colrev_status"]
                == colrev.record.RecordState.pdf_needs_manual_preparation
            ):
                record = colrev.record.Record(data=record_dict)
                record.set_status(
                    target_state=colrev.record.RecordState.pdf_not_available
                )
        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        self.review_manager.create_commit(
            msg="Discard man-prep PDFs", manual_author=True
        )

    def get_data(self) -> dict:
        """Get the data for PDF prep man"""
        # pylint: disable=duplicate-code

        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        nr_tasks = len(
            [
                x
                for x in record_header_list
                if colrev.record.RecordState.pdf_needs_manual_preparation
                == x["colrev_status"]
            ]
        )
        pad = 0
        if record_header_list:
            pad = min((max(len(x["ID"]) for x in record_header_list) + 2), 40)

        items = self.review_manager.dataset.read_next_record(
            conditions=[
                {
                    "colrev_status": colrev.record.RecordState.pdf_needs_manual_preparation
                }
            ]
        )
        pdf_prep_man_data = {"nr_tasks": nr_tasks, "PAD": pad, "items": items}
        self.review_manager.logger.debug(
            self.review_manager.p_printer.pformat(pdf_prep_man_data)
        )
        return pdf_prep_man_data

    def pdfs_prepared_manually(self) -> bool:
        """Check whether PDFs were prepared manually"""
        return self.review_manager.dataset.has_changes()

    def pdf_prep_man_stats(self) -> None:
        """Determine PDF prep man statistics"""
        # pylint: disable=duplicate-code

        self.review_manager.logger.info(
            f"Load {self.review_manager.dataset.RECORDS_FILE_RELATIVE}"
        )
        records = self.review_manager.dataset.load_records_dict()

        self.review_manager.logger.info("Calculate statistics")
        stats: dict = {"ENTRYTYPE": {}}

        prep_man_hints = []
        crosstab = []
        for record_dict in records.values():
            if (
                colrev.record.RecordState.pdf_needs_manual_preparation
                != record_dict["colrev_status"]
            ):
                continue

            if record_dict["ENTRYTYPE"] in stats["ENTRYTYPE"]:
                stats["ENTRYTYPE"][record_dict["ENTRYTYPE"]] = (
                    stats["ENTRYTYPE"][record_dict["ENTRYTYPE"]] + 1
                )
            else:
                stats["ENTRYTYPE"][record_dict["ENTRYTYPE"]] = 1

            record = colrev.record.Record(data=record_dict)
            prov_d = record.data["colrev_data_provenance"]

            if "file" in prov_d:
                if prov_d["file"]["note"] != "":
                    for hint in prov_d["file"]["note"].split(","):
                        prep_man_hints.append(hint.lstrip())

            for hint in prep_man_hints:
                crosstab.append([record_dict["journal"], hint.lstrip()])

        crosstab_df = pd.DataFrame(crosstab, columns=["journal", "hint"])

        if crosstab_df.empty:
            print("No records to prepare manually.")
        else:
            # pylint: disable=duplicate-code
            tabulated = pd.pivot_table(
                crosstab_df[["journal", "hint"]],
                index=["journal"],
                columns=["hint"],
                aggfunc=len,
                fill_value=0,
                margins=True,
            )
            # .sort_index(axis='columns')
            tabulated.sort_values(by=["All"], ascending=False, inplace=True)
            # Transpose because we tend to have more error categories than search files.
            tabulated = tabulated.transpose()
            print(tabulated)
            self.review_manager.logger.info(
                "Writing data to file: manual_preparation_statistics.csv"
            )
            tabulated.to_csv("manual_pdf_preparation_statistics.csv")

    def extract_needs_pdf_prep_man(self) -> None:
        """Apply PDF prep man to csv/bib"""

        prep_bib_path = self.review_manager.path / Path("data/pdf-prep-records.bib")
        prep_csv_path = self.review_manager.path / Path("data/pdf-prep-records.csv")

        if prep_csv_path.is_file():
            print(f"Please rename file to avoid overwriting changes ({prep_csv_path})")
            return

        if prep_bib_path.is_file():
            print(f"Please rename file to avoid overwriting changes ({prep_bib_path})")
            return

        self.review_manager.logger.info(
            f"Load {self.review_manager.dataset.RECORDS_FILE_RELATIVE}"
        )
        records = self.review_manager.dataset.load_records_dict()

        records = {
            record["ID"]: record
            for record in records.values()
            if colrev.record.RecordState.pdf_needs_manual_preparation
            == record["colrev_status"]
        }
        self.review_manager.dataset.save_records_dict_to_file(
            records=records, save_path=prep_bib_path
        )

        bib_db_df = pd.DataFrame.from_records(list(records.values()))

        # pylint: disable=duplicate-code
        col_names = [
            "ID",
            "colrev_origin",
            "author",
            "title",
            "year",
            "journal",
            # "booktitle",
            "volume",
            "number",
            "pages",
            "doi",
        ]
        for col_name in col_names:
            if col_name not in bib_db_df:
                bib_db_df[col_name] = "NA"
        bib_db_df = bib_db_df[col_names]

        bib_db_df.to_csv(prep_csv_path, index=False)
        self.review_manager.logger.info(f"Created {prep_csv_path.name}")

    def apply_pdf_prep_man(self) -> None:
        """Apply PDF prep man from csv/bib"""

        if Path("data/pdf-prep-records.csv").is_file():
            self.review_manager.logger.info("Load prep-records.csv")
            bib_db_df = pd.read_csv("data/pdf-prep-records.csv")
            records_changed = bib_db_df.to_dict("records")

        if Path("data/pdf-prep-records.bib").is_file():
            self.review_manager.logger.info("Load prep-records.bib")

            with open("data/pdf-prep-records.bib", encoding="utf8") as target_db:
                records_changed_dict = self.review_manager.dataset.load_records_dict(
                    load_str=target_db.read()
                )
                records_changed = list(records_changed_dict.values())

        records = self.review_manager.dataset.load_records_dict()
        for record in records.values():
            # IDs may change - matching based on origins
            changed_record_l = [
                x
                for x in records_changed
                if x["colrev_origin"] == record["colrev_origin"]
            ]
            if len(changed_record_l) == 1:
                changed_record = changed_record_l.pop()
                for key, value in changed_record.items():
                    # if record['ID'] == 'Alter2014':
                    #     print(key, value)
                    if str(value) == "nan":
                        if key in record:
                            del record[key]
                        continue
                    record[key] = value
                    if value == "":
                        del record[key]

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.check_repo()

    def extract_coverpage(self, *, filepath: Path) -> None:
        """Extract coverpage from PDF"""

        local_index = self.review_manager.get_local_index()
        cp_path = local_index.local_environment_path / Path(".coverpages")
        cp_path.mkdir(exist_ok=True)

        try:
            pdf_reader = PdfFileReader(str(filepath), strict=False)
            writer_cp = PdfFileWriter()
            writer_cp.addPage(pdf_reader.getPage(0))
            writer = PdfFileWriter()
            for i in range(1, len(pdf_reader.pages)):
                writer.addPage(pdf_reader.getPage(i))
            with open(filepath, "wb") as outfile:
                writer.write(outfile)
            with open(cp_path / filepath.name, "wb") as outfile:
                writer_cp.write(outfile)
        except PdfReadError as exc:
            raise colrev_exceptions.InvalidPDFException(filepath) from exc

    def extract_lastpage(self, *, filepath: Path) -> None:
        """Extract last page from PDF"""

        local_index = self.review_manager.get_local_index()
        lp_path = local_index.local_environment_path / Path(".lastpages")
        lp_path.mkdir(exist_ok=True)

        try:
            pdf_reader = PdfFileReader(str(filepath), strict=False)
            writer_lp = PdfFileWriter()
            writer_lp.addPage(pdf_reader.getPage(len(pdf_reader.pages) - 1))
            writer = PdfFileWriter()
            for i in range(0, len(pdf_reader.pages) - 1):
                writer.addPage(pdf_reader.getPage(i))
            with open(filepath, "wb") as outfile:
                writer.write(outfile)
            with open(lp_path / filepath.name, "wb") as outfile:
                writer_lp.write(outfile)
        except PdfReadError as exc:
            raise colrev_exceptions.InvalidPDFException(filepath) from exc

    def extract_pages(self, *, filepath: Path, pages_to_remove: list) -> None:
        """Extract pages from PDF"""

        try:
            pdf_reader = PdfFileReader(str(filepath), strict=False)
            pages_to_add = [
                x
                for x in list(range(0, len(pdf_reader.pages)))
                if x not in pages_to_remove
            ]
            writer = PdfFileWriter()
            for i in pages_to_add:
                writer.addPage(pdf_reader.getPage(i))
            with open(filepath, "wb") as outfile:
                writer.write(outfile)

        except PdfReadError as exc:
            raise colrev_exceptions.InvalidPDFException(filepath) from exc

    def set_pdf_man_prepared(self, *, record: colrev.record.Record) -> None:
        """Set the PDF to manually prepared"""

        record.set_status(target_state=colrev.record.RecordState.pdf_prepared)
        record.reset_pdf_provenance_notes()

        pdf_path = Path(self.review_manager.path / Path(record.data["file"]))
        prev_cpid = record.data.get("colrev_pdf_id", "NA")
        record.data.update(colrev_pdf_id=record.get_colrev_pdf_id(pdf_path=pdf_path))
        if prev_cpid != record.data.get("colrev_pdf_id", "NA"):
            record.add_data_provenance(key="file", source="manual")

        record_dict = record.get_data()
        self.review_manager.dataset.save_records_dict(
            records={record_dict["ID"]: record_dict}, partial=True
        )
        self.review_manager.dataset.add_changes(
            path=self.review_manager.dataset.RECORDS_FILE_RELATIVE
        )

    @colrev.operation.Operation.decorate()
    def main(self) -> None:
        """Prepare PDFs manually (main entrypoint)"""

        if self.review_manager.in_ci_environment():
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="colrev pdf-prep-man",
                detailed_trace="pdf-prep-man not available in ci environment",
            )

        records = self.review_manager.dataset.load_records_dict()

        package_manager = self.review_manager.get_package_manager()

        pdf_prep_man_package_endpoints = (
            self.review_manager.settings.pdf_prep.pdf_prep_man_package_endpoints
        )
        for pdf_prep_man_package_endpoint in pdf_prep_man_package_endpoints:
            endpoint_dict = package_manager.load_packages(
                package_type=colrev.env.package_manager.PackageEndpointType.pdf_prep_man,
                selected_packages=pdf_prep_man_package_endpoints,
                operation=self,
            )

            endpoint = endpoint_dict[pdf_prep_man_package_endpoint["endpoint"]]

            records = endpoint.pdf_prep_man(self, records)  # type: ignore



#!/usr/bin/env python3
"""Create a commit, including the CoLRev report."""
from __future__ import annotations

import importlib
import os
import shutil
import sys
import tempfile
from importlib.metadata import version
from pathlib import Path
from typing import Optional
from typing import TYPE_CHECKING

import git

import colrev.env.utils
import colrev.exceptions as colrev_exceptions

if TYPE_CHECKING:
    import colrev.review_manager


class Commit:
    """Create commits"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods

    __temp_path = Path.home().joinpath("colrev") / Path(".colrev_temp")

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        msg: str,
        manual_author: bool,
        script_name: str,
        saved_args: Optional[dict] = None,
    ) -> None:
        self.review_manager = review_manager
        self.manual_author = manual_author

        self.msg = msg
        self.script_name = self.__parse_script_name(script_name=script_name)
        self.saved_args = self.__parse_saved_args(saved_args=saved_args)
        self.tree_hash = ""

        self.last_commit_sha = ""
        try:
            self.last_commit_sha = review_manager.dataset.get_last_commit_sha()
        except ValueError:
            pass

        self.records_committed = review_manager.dataset.records_file.is_file()
        self.completeness_condition = review_manager.get_completeness_condition()
        self.colrev_version = f'version {version("colrev")}'
        sys_v = sys.version
        self.python_version = f'version {sys_v[: sys_v.find(" ")]}'
        stream = os.popen("git --version")
        self.git_version = stream.read().replace("git ", "").replace("\n", "")
        stream = os.popen("docker --version")
        self.docker_version = stream.read().replace("Docker ", "").replace("\n", "")
        if self.docker_version == "":
            self.docker_version = "Not installed"

        self.ext_script_name = ""
        self.ext_script_version = ""

        if script_name != "":
            # Note : the script_name / ext_script seems to be different on macos during init!?
            ext_script = script_name.split(" ")[0]
            if ext_script != "colrev":
                try:
                    script_version = version(ext_script)
                    self.ext_script_name = script_name
                    self.ext_script_version = f"version {script_version}"
                except importlib.metadata.PackageNotFoundError:
                    pass
                except ValueError:
                    # Note : macos error in importlib.metadata.version
                    # ValueError: A distribution name is required
                    self.ext_script_name = "unknown"
                    self.ext_script_version = "unknown"

        self.__temp_path.mkdir(exist_ok=True, parents=True)

    def __parse_saved_args(self, *, saved_args: Optional[dict] = None) -> str:
        saved_args_str = ""
        if saved_args is not None:
            for key, value in saved_args.items():
                if isinstance(value, (bool, float, int, str)):
                    if value == "":
                        saved_args_str = saved_args_str + f"     --{key} \\\n"
                    else:
                        saved_args_str = saved_args_str + f"     --{key}={value} \\\n"
            # Replace the last backslash (for argument chaining across linebreaks)
            saved_args_str = saved_args_str.rstrip(" \\\n")

        return saved_args_str

    def __parse_script_name(self, *, script_name: str) -> str:
        if script_name == "MANUAL":
            script_name = "Commit created manually or by external script"
        elif " " in script_name:
            script_name = script_name.replace("colrev cli", "colrev").replace(
                "prescreen_cli", "prescreen"
            )
        return script_name

    def __get_version_flag(self) -> str:
        flag = ""
        if "dirty" in version("colrev"):
            flag = "*"
        return flag

    def __get_commit_report_header(self) -> str:
        template = colrev.env.utils.get_template(
            template_path="template/ops/commit_report_header.txt"
        )
        content = template.render(commit_details=self)

        return content

    def __get_commit_report_details(self) -> str:
        template = colrev.env.utils.get_template(
            template_path="template/ops/commit_report_details.txt"
        )
        content = template.render(commit_details=self)

        return content

    def __get_detailed_processing_report(self) -> str:
        processing_report = ""
        if self.review_manager.report_path.is_file():
            # Reformat
            prefixes = [
                "[('change', 'author',",
                "[('change', 'title',",
                "[('change', 'journal',",
                "[('change', 'booktitle',",
            ]

            with tempfile.NamedTemporaryFile(
                dir=self.__temp_path, mode="r+b", delete=False
            ) as temp:
                with open(self.review_manager.report_path, "r+b") as file:
                    shutil.copyfileobj(file, temp)  # type: ignore
            # self.report_path.rename(temp.name)
            with open(temp.name, encoding="utf8") as reader, open(
                self.review_manager.report_path, "w", encoding="utf8"
            ) as writer:
                line = reader.readline()
                while line:
                    if (
                        any(prefix in line for prefix in prefixes)
                        and "', '" in line[30:]
                    ):
                        split_pos = line.rfind("', '") + 2
                        indent = line.find("', (") + 3
                        writer.write(line[:split_pos] + "\n")
                        writer.write(" " * indent + line[split_pos:])
                    else:
                        writer.write(line)

                    line = reader.readline()

            with open(self.review_manager.report_path, encoding="utf8") as file:
                line = file.readline()
                debug_part = False
                while line:
                    # For more efficient debugging (loading of dict with Enum)
                    if "colrev_status" in line and "<RecordState." in line:
                        line = line.replace("<RecordState", "RecordState")
                        line = line[: line.rfind(":")] + line[line.rfind(">") + 1 :]
                    if "[DEBUG]" in line or debug_part:
                        debug_part = True
                        if any(
                            x in line
                            for x in ["[INFO]", "[ERROR]", "[WARNING]", "[CRITICAL"]
                        ):
                            debug_part = False
                    if not debug_part:
                        processing_report = processing_report + line
                    line = file.readline()

            processing_report = "\nProcessing report\n" + "".join(processing_report)
        return processing_report

    def __get_commit_report(self) -> str:
        status_operation = self.review_manager.get_status_operation()

        report = self.__get_commit_report_header()
        report += status_operation.get_review_status_report(colors=False)
        report += self.__get_commit_report_details()

        return report

    def create(self, *, skip_status_yaml: bool = False) -> bool:
        """Create a commit (including the commit message and details)"""

        if self.review_manager.dataset.has_changes():
            self.review_manager.logger.debug("Prepare commit: checks and updates")
            if not skip_status_yaml:
                self.review_manager.update_status_yaml()
                self.review_manager.dataset.add_changes(
                    path=self.review_manager.STATUS_RELATIVE
                )

            committer, email = self.review_manager.get_committer()

            if self.manual_author:
                git_author = git.Actor(committer, email)
            else:
                git_author = git.Actor(f"script:{self.script_name}", email)

            # Note : this should run as the last command before creating the commit
            # to ensure that the git tree_hash is up-to-date.
            self.tree_hash = self.review_manager.dataset.get_tree_hash()
            self.msg = (
                self.msg
                + self.__get_version_flag()
                + self.__get_commit_report()
                + self.__get_detailed_processing_report()
            )
            self.review_manager.dataset.create_commit(
                msg=self.msg,
                author=git_author,
                committer=git.Actor(committer, email),
                hook_skipping=True,
            )

            self.review_manager.logger.info("Created commit")
            self.review_manager.reset_report_logger()
            if self.review_manager.dataset.has_changes():
                raise colrev_exceptions.DirtyRepoAfterProcessingError(
                    "A clean repository is expected."
                )
            return True

        return False

    def update_report(self, *, msg_file: Path) -> None:
        """Update the report"""
        report = self.__get_commit_report()
        with open(msg_file, "a", encoding="utf8") as file:
            file.write(report)
